# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from typing import Any
from typing import Dict
from typing import Union

import pandas as pd
import pandas.testing as pdt
from pandas.core.algorithms import safe_sort

from kats.compat import compat

version: compat.Version = compat.Version("pandas")


def convert_precision(check_less_precise: Union[bool, int], rtol: float) -> int:
    """Convert rtol and check_less_precise argument."""
    if check_less_precise is True:
        precise = 3
    elif check_less_precise is False:
        precise = 5
    else:
        precise = check_less_precise
    precise = min(precise, int(abs(math.log10(rtol))))
    return precise


def assert_frame_equal(
    left: pd.DataFrame,
    right: pd.DataFrame,
    check_dtype: bool = True,
    check_index_type: Union[bool, str] = "equiv",
    check_column_type: Union[bool, str] = "equiv",
    check_frame_type: bool = True,
    check_less_precise: Union[bool, int] = False,
    check_names: bool = True,
    by_blocks: bool = False,
    check_exact: bool = False,
    check_datetimelike_compat: bool = False,
    check_categorical: bool = True,
    check_like: bool = False,
    check_freq: bool = True,
    check_flags: bool = True,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    obj: str = "DataFrame",
) -> None:
    kwargs: Dict[str, Any] = {
        "check_dtype": check_dtype,
        "check_index_type": check_index_type,
        "check_column_type": check_column_type,
        "check_frame_type": check_frame_type,
        "check_names": check_names,
        "by_blocks": by_blocks,
        "check_exact": check_exact,
        "check_datetimelike_compat": check_datetimelike_compat,
        "check_categorical": check_categorical,
        "check_like": check_like,
        "obj": obj,
    }
    if version < "1.1":
        kwargs["check_less_precise"] = convert_precision(check_less_precise, rtol)
    else:
        kwargs["check_freq"] = check_freq
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
        if version >= "1.2":
            kwargs["check_flags"] = check_flags
    pdt.assert_frame_equal(left, right, **kwargs)
    if (
        check_freq
        and version < "1.1"
        and isinstance(left.index, (pd.DatetimeIndex, pd.TimedeltaIndex))
        and hasattr(left.index, "freq")
        and hasattr(right.index, "freq")
    ):
        assert left.index.freq == right.index.freq, (left.index.freq, right.index.freq)
    if check_flags and version < "1.2" and hasattr(left, "flags") and hasattr(right, "flags"):
        assert left.flags == right.flags, f"{repr(left.flags)} != {repr(right.flags)}"


def assert_series_equal(
    left: pd.Series,
    right: pd.Series,
    check_dtype: bool = True,
    check_index_type: Union[bool, str] = "equiv",
    check_series_type: Union[bool, str] = "equiv",
    check_less_precise: Union[bool, int] = False,
    check_names: bool = True,
    check_exact: bool = False,
    check_datetimelike_compat: bool = False,
    check_categorical: bool = True,
    check_category_order: bool = False,
    check_freq: bool = True,
    check_flags: bool = True,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    obj: str = "Series",
    check_index: bool = True,
) -> None:
    kwargs: Dict[str, Any] = {
        "check_dtype": check_dtype,
        "check_index_type": check_index_type,
        "check_series_type": check_series_type,
        "check_names": check_names,
        "check_exact": check_exact,
        "check_datetimelike_compat": check_datetimelike_compat,
        "check_categorical": check_categorical,
        "obj": obj,
    }
    if version < "1.0.2":
        kwargs["check_less_precise"] = convert_precision(check_less_precise, rtol)
    elif version < "1.1":
        kwargs["check_less_precise"] = convert_precision(check_less_precise, rtol)
        kwargs["check_category_order"] = check_category_order
    elif version < "1.2":
        kwargs["check_category_order"] = check_category_order
        kwargs["check_freq"] = check_freq
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
    elif version < "1.3":
        kwargs["check_category_order"] = check_category_order
        kwargs["check_freq"] = check_freq
        kwargs["check_flags"] = check_flags
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
    else:
        kwargs["check_category_order"] = check_category_order
        kwargs["check_freq"] = check_freq
        kwargs["check_flags"] = check_flags
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
        kwargs["check_index"] = check_index

    pdt.assert_series_equal(left, right, **kwargs)
    if (
        check_freq
        and version < "1.1"
        and isinstance(left.index, (pd.DatetimeIndex, pd.TimedeltaIndex))
        and hasattr(left.index, "freq")
        and hasattr(right.index, "freq")
    ):
        # pyre-fixme[16]: `Index` has no attribute `freq`.
        assert left.index.freq == right.index.freq, (left.index.freq, right.index.freq)
    if check_flags and version < "1.2" and hasattr(left, "flags") and hasattr(right, "flags"):
        assert left.flags == right.flags, f"{repr(left.flags)} != {repr(right.flags)}"
    if check_index and version < "1.3":
        assert_index_equal(
            left.index,
            right.index,
            exact=check_index_type,
            check_names=check_names,
            check_less_precise=check_less_precise,
            check_exact=check_exact,
            check_categorical=check_categorical,
            rtol=rtol,
            atol=atol,
            obj=f"{obj}.index",
        )


def assert_index_equal(
    left: pd.Index,
    right: pd.Index,
    exact: Union[bool, str] = "equiv",
    check_names: bool = True,
    check_less_precise: Union[bool, int] = False,
    check_exact: bool = True,
    check_categorical: bool = True,
    check_order: bool = True,
    rtol: float = 1e-05,
    atol: float = 1e-08,
    obj: str = "Index",
) -> None:
    if not check_order and version < "1.2":
        left = pd.Index(safe_sort(left))
        right = pd.Index(safe_sort(right))

    kwargs: Dict[str, Any] = {
        "exact": exact,
        "check_names": check_names,
        "check_exact": check_exact,
        "check_categorical": check_categorical,
        "obj": obj,
    }
    if version < "1.1":
        kwargs["check_less_precise"] = convert_precision(check_less_precise, rtol)
    elif version < "1.2":
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
    else:
        kwargs["check_order"] = check_order
        kwargs["rtol"] = rtol
        kwargs["atol"] = atol
    pdt.assert_index_equal(left, right, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from operator import attrgetter
from typing import Optional
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized
from scipy.stats import chi2  # @manual
from sklearn.datasets import make_spd_matrix

from kats.consts import TimeSeriesData
from kats.detectors.cusum_detection import CUSUMDetector
from kats.detectors.cusum_detection import MultiCUSUMDetector
from kats.detectors.cusum_detection import VectorizedCUSUMDetector


class CUSUMDetectorTest(TestCase):
    def setUp(self) -> None:
        np.random.seed(10)

        # increasing with variance detection setup
        df_increase = pd.DataFrame(
            {"increase": np.concatenate([np.random.normal(1, 0.2, 30), np.random.normal(1.5, 0.2, 30)])}
        )

        df_increase["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))
        inc_timeseries = TimeSeriesData(df_increase)
        self.inc_detector = CUSUMDetector(inc_timeseries)
        self.inc_change_points = self.inc_detector.detector()
        self.inc_metadata = self.inc_change_points[0]

        # decreasing detection setup
        df_decrease = pd.DataFrame(
            {"decrease": np.concatenate([np.random.normal(1, 0.2, 50), np.random.normal(0.5, 0.2, 10)])}
        )

        df_decrease["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        dec_timeseries = TimeSeriesData(df_decrease)
        self.dec_detector = CUSUMDetector(dec_timeseries)
        self.dec_change_points = self.dec_detector.detector()
        self.dec_metadata = self.dec_change_points[0]

        # seasonality setup
        self.periodicity = 48
        self.total_cycles = 3
        harmonics = 2
        noise_std = 3

        seasonal_term = CUSUMDetectorTest.simulate_seasonal_term(
            self.periodicity,
            self.total_cycles,
            noise_std=noise_std,
            harmonics=harmonics,
        )
        seasonal_term = seasonal_term / seasonal_term.std() * 2
        residual = np.random.normal(0, 1, self.periodicity * self.total_cycles)
        self.seasonal_data = seasonal_term + residual

        # seasonality with increase trend setup
        trend_term = np.logspace(0, 1, self.periodicity * self.total_cycles)
        data = self.seasonal_data + trend_term
        data -= np.min(data)

        df_seasonality = pd.DataFrame(
            {
                "time": pd.date_range(
                    "2020-01-01",
                    periods=self.periodicity * self.total_cycles,
                    freq="30T",
                ),
                "seasonality": data,
            }
        )
        timeseries = TimeSeriesData(df_seasonality)
        self.season_inc_trend_detector = CUSUMDetector(timeseries)
        self.season_inc_trend_change_points = self.season_inc_trend_detector.detector(
            interest_window=[
                self.periodicity * (self.total_cycles - 1),
                self.periodicity * self.total_cycles - 1,
            ],
            magnitude_quantile=1,
            change_directions=["increase", "decrease"],
            delta_std_ratio=0,
        )
        self.season_metadata = self.season_inc_trend_change_points[0]

        # test on step change with no variance
        df_increase_no_var = pd.DataFrame(
            {"increase": np.concatenate([np.random.normal(1, 0, 30), np.random.normal(2, 0, 30)])}
        )

        df_increase_no_var["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        no_var_timeseries = TimeSeriesData(df_increase_no_var)
        self.no_var_detector = CUSUMDetector(no_var_timeseries)
        self.no_var_change_points = self.no_var_detector.detector()

        # no seasonality setup
        data = self.seasonal_data
        data -= np.min(data)

        df_seasonality = pd.DataFrame(
            {
                "time": pd.date_range(
                    "2020-01-01",
                    periods=self.periodicity * self.total_cycles,
                    freq="30T",
                ),
                "seasonality": data,
            }
        )
        timeseries = TimeSeriesData(df_seasonality)
        self.no_season_detector = CUSUMDetector(timeseries)
        self.no_season_change_points = self.no_season_detector.detector(
            interest_window=[
                self.periodicity * (self.total_cycles - 1),
                self.periodicity * self.total_cycles - 1,
            ],
            magnitude_quantile=1,
            change_directions=["increase"],
            delta_std_ratio=0,
        )

        # no regression setup
        df_noregress = pd.DataFrame({"no_change": np.random.normal(1, 0.2, 60)})

        df_noregress["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        timeseries = TimeSeriesData(df_noregress)
        self.no_reg_detector = CUSUMDetector(timeseries)
        self.no_reg_change_points = self.no_reg_detector.detector(start_point=20)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["inc_change_points", 1],
            ["dec_change_points", 1],
            ["season_inc_trend_change_points", 1],
            ["no_var_change_points", 1],
            ["no_reg_change_points", 0],
            ["no_season_change_points", 0],
        ]
    )
    def test_cp_len(self, cp_name: str, expected: int) -> None:
        self.assertEqual(len(attrgetter(cp_name)(self)), expected)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata", 29],
    #  ["dec_metadata", 49]])`.
    @parameterized.expand(
        [
            ["inc_metadata", 29],
            ["dec_metadata", 49],
        ]
    )
    def test_cp_index(self, metadata_name: str, expected: int) -> None:
        self.assertLessEqual(abs(attrgetter(metadata_name)(self).cp_index - expected), 1)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata", "increase"],
    #  ["dec_metadata", "decrease"]])`.
    @parameterized.expand(
        [
            ["inc_metadata", "increase"],
            ["dec_metadata", "decrease"],
        ]
    )
    def test_direction(self, metadata_name: str, expected: str) -> None:
        self.assertEqual(attrgetter(metadata_name)(self).direction, expected)

    def test_increasing_mu(self) -> None:
        self.assertLess(self.inc_metadata.mu0, self.inc_metadata.mu1)

    def test_increasing_correct_delta(self) -> None:
        self.assertEqual(self.inc_metadata.delta, self.inc_metadata.mu1 - self.inc_metadata.mu0)

    def test_increasing_regression(self) -> None:
        self.assertTrue(self.inc_metadata.regression_detected)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["season_metadata.p_value_int", "season_metadata.llr_int"],
            ["inc_metadata.p_value", "inc_metadata.llr"],
        ]
    )
    def test_p_val(self, pval_name: str, llr_name: str) -> None:
        self.assertEqual(
            attrgetter(pval_name)(self),
            1 - chi2.cdf(attrgetter(llr_name)(self), 2),
        )

    def test_increasing_p_val_nan(self) -> None:
        self.assertTrue(np.isnan(self.inc_metadata.p_value_int))

    def test_increasing_llr_int(self) -> None:
        self.assertEqual(self.inc_metadata.llr_int, np.inf)

    def test_increasing_stable_changepoint(self) -> None:
        self.assertTrue(self.inc_metadata.stable_changepoint)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["inc_detector", "inc_change_points"],
            ["dec_detector", "dec_change_points"],
            ["season_inc_trend_detector", "season_inc_trend_change_points"],
            ["no_var_detector", "no_var_change_points"],
            ["no_reg_detector", "no_reg_change_points"],
            ["no_season_detector", "no_season_change_points"],
        ]
    )
    def test_plot(self, detector_name: str, cp_name: str) -> None:
        attrgetter(detector_name)(self).plot(attrgetter(cp_name)(self))

    @staticmethod
    def simulate_seasonal_term(
        periodicity: int,
        total_cycles: int,
        noise_std: float = 1.0,
        harmonics: Optional[float] = None,
    ) -> np.ndarray:
        duration = periodicity * total_cycles
        assert duration == int(duration)
        duration = int(duration)
        harmonics = harmonics if harmonics else int(np.floor(periodicity / 2))

        lambda_p = 2 * np.pi / float(periodicity)

        gamma_jt = noise_std * np.random.randn((harmonics))
        gamma_star_jt = noise_std * np.random.randn((harmonics))

        total_timesteps = 100 * duration  # Pad for burn in
        series = np.zeros(total_timesteps)
        for t in range(total_timesteps):
            gamma_jtp1 = np.zeros_like(gamma_jt)
            gamma_star_jtp1 = np.zeros_like(gamma_star_jt)
            # pyre-ignore[6]: For 2nd param expected `SupportsIndex` but got `float`.
            for j in range(1, harmonics + 1):
                cos_j = np.cos(lambda_p * j)
                sin_j = np.sin(lambda_p * j)
                gamma_jtp1[j - 1] = (
                    gamma_jt[j - 1] * cos_j + gamma_star_jt[j - 1] * sin_j + noise_std * np.random.randn()
                )
                gamma_star_jtp1[j - 1] = (
                    -gamma_jt[j - 1] * sin_j + gamma_star_jt[j - 1] * cos_j + noise_std * np.random.randn()
                )
            series[t] = np.sum(gamma_jtp1)
            gamma_jt = gamma_jtp1
            gamma_star_jt = gamma_star_jtp1
        wanted_series = series[-duration:]  # Discard burn in

        return wanted_series

    def test_seasonality_with_increasing_trend_cp_index(self) -> None:
        self.assertGreaterEqual(
            self.season_metadata.cp_index,
            # pyre-fixme[6]: For 2nd param expected `SupportsDunderLE[Variable[_T]]`
            #  but got `int`.
            self.periodicity * (self.total_cycles - 1),
        )

    def test_logging_multivariate_error(self) -> None:
        # test multivariate error
        np.random.seed(10)
        df_multi_var = pd.DataFrame(
            {
                "no_change": np.random.normal(1, 0.2, 60),
                "no_change2": np.random.normal(1, 0.2, 60),
            }
        )

        df_multi_var["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        with self.assertRaises(ValueError):
            timeseries = TimeSeriesData(df_multi_var)
            CUSUMDetector(timeseries)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["WARNING", 0.900000],
    #  ["DEBUG", None]])`.
    @parameterized.expand(
        [
            ["WARNING", 0.9],
            ["DEBUG", None],
        ]
    )
    # pyre-ignore[2]: Parameter must be annotated.
    def test_logging_neg_magnitude(self, level, mag_q) -> None:
        # test logging setup - negative in magnitude
        np.random.seed(10)
        df_neg = pd.DataFrame({"no_change": -np.random.normal(1, 0.2, 60)})

        df_neg["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        timeseries = TimeSeriesData(df_neg)
        logging_detector = CUSUMDetector(timeseries)

        with self.assertLogs(level=level):
            logging_detector.detector(magnitude_quantile=mag_q, interest_window=[40, 60])

    def test_ts_without_name(self) -> None:
        n = 10
        time = pd.Series(pd.date_range(start="2018-01-01", periods=n, freq="D"))
        value = pd.Series(np.arange(n))
        ts = TimeSeriesData(time=time, value=value)

        detector = CUSUMDetector(ts)
        change_points = detector.detector()
        detector.plot(change_points)


class MultiCUSUMDetectorTest(TestCase):
    def setUp(self) -> None:
        # increasing setup
        self.D = 10
        random_state = 10
        np.random.seed(random_state)
        mean1 = np.ones(self.D)
        mean2 = mean1 * 2
        sigma = make_spd_matrix(self.D, random_state=random_state)

        df_increase = pd.DataFrame(
            np.concatenate(
                [
                    np.random.multivariate_normal(mean1, sigma, 60),
                    np.random.multivariate_normal(mean2, sigma, 30),
                ]
            )
        )

        df_increase["time"] = pd.Series(pd.date_range("2019-01-01", "2019-04-01"))

        timeseries_increase = TimeSeriesData(df_increase)
        self.inc_change_points = MultiCUSUMDetector(timeseries_increase).detector()
        self.inc_metadata = self.inc_change_points[0]

        # decreasing setup
        df_decrease = pd.DataFrame(
            np.concatenate(
                [
                    np.random.multivariate_normal(mean2, sigma, 60),
                    np.random.multivariate_normal(mean1, sigma, 30),
                ]
            )
        )

        df_decrease["time"] = pd.Series(pd.date_range("2019-01-01", "2019-04-01"))

        timeseries_decrease = TimeSeriesData(df_decrease)
        self.dec_change_points = MultiCUSUMDetector(timeseries_decrease).detector()
        self.dec_metadata = self.dec_change_points[0]

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_change_points"],
    #  ["dec_change_points"]])`.
    @parameterized.expand(
        [
            ["inc_change_points"],
            ["dec_change_points"],
        ]
    )
    def test_cp_len(self, cp_name: str) -> None:
        self.assertEqual(len(attrgetter(cp_name)(self)), 1)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_cp_index(self, cp_name: str) -> None:
        self.assertLessEqual(abs(attrgetter(cp_name)(self).cp_index - 59), 1)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata.mu0",
    #  "inc_metadata.mu1"], ["dec_metadata.mu1", "dec_metadata.mu0"]])`.
    @parameterized.expand(
        [
            ["inc_metadata.mu0", "inc_metadata.mu1"],
            ["dec_metadata.mu1", "dec_metadata.mu0"],
        ]
    )
    def test_mu(self, m1_name: str, m2_name: str) -> None:
        for m1, m2 in zip(attrgetter(m1_name)(self), attrgetter(m2_name)(self)):
            self.assertLess(m1, m2)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata",
    #  "inc_metadata.mu0", "inc_metadata.mu1"], ["dec_metadata", "dec_metadata.mu0",
    #  "dec_metadata.mu1"]])`.
    @parameterized.expand(
        [
            ["inc_metadata", "inc_metadata.mu0", "inc_metadata.mu1"],
            ["dec_metadata", "dec_metadata.mu0", "dec_metadata.mu1"],
        ]
    )
    def test_correct_delta(self, metadata_name: str, mu0_name: str, mu1_name: str) -> None:
        for d, diff in zip(
            attrgetter(metadata_name)(self).delta,
            attrgetter(mu1_name)(self) - attrgetter(mu0_name)(self),
        ):
            self.assertEqual(d, diff)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_regression(self, metadata_name: str) -> None:
        self.assertTrue(attrgetter(metadata_name)(self).regression_detected)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_p_val(self, metadata_name: str) -> None:
        self.assertEqual(
            attrgetter(metadata_name)(self).p_value,
            1 - chi2.cdf(attrgetter(metadata_name)(self).llr, self.D + 1),
        )

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_gaussian_increase_p_val_nan(self, metadata_name: str) -> None:
        self.assertTrue(np.isnan(attrgetter(metadata_name)(self).p_value_int))

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_gaussian_increase_llr_int(self, metadata_name: str) -> None:
        self.assertEqual(attrgetter(metadata_name)(self).llr_int, np.inf)

    # pyre-ignore[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["inc_metadata"],
    #  ["dec_metadata"]])`.
    @parameterized.expand(
        [
            ["inc_metadata"],
            ["dec_metadata"],
        ]
    )
    def test_gaussian_increase_stable_changepoint(self, metadata_name: str) -> None:
        self.assertTrue(attrgetter(metadata_name)(self).stable_changepoint)

    def test_no_changepoint(self) -> None:
        D = 10
        random_state = 10
        np.random.seed(random_state)
        mean = np.ones(D)
        sigma = make_spd_matrix(D, random_state=random_state)
        # Use the same mean for the entire series and there should be no changepoint
        df_no_change = pd.DataFrame(np.random.multivariate_normal(mean, sigma, 90))
        df_no_change["time"] = pd.Series(pd.date_range("2019-01-01", "2019-04-01"))

        timeseries_no_change = TimeSeriesData(df_no_change)
        change_points = MultiCUSUMDetector(timeseries_no_change).detector()
        self.assertEqual(len(change_points), 0)


class VectorizedCUSUMDetectorTest(TestCase):
    def setUp(self) -> None:
        np.random.seed(10)

        # increasing with variance detection setup
        df = pd.DataFrame(
            {
                "increase": np.concatenate([np.random.normal(1, 0.2, 30), np.random.normal(1.5, 0.2, 30)]),
                "decrease": np.concatenate([np.random.normal(1, 0.2, 50), np.random.normal(0.5, 0.2, 10)]),
            }
        )
        df["time"] = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        self.inc_change_points = CUSUMDetector(TimeSeriesData(df[["increase", "time"]])).detector()
        self.dec_change_points = CUSUMDetector(TimeSeriesData(df[["decrease", "time"]])).detector()
        self.dec_change_points_int_window = CUSUMDetector(TimeSeriesData(df[["decrease", "time"]])).detector(
            change_directions=["decrease"], interest_window=(35, 55)
        )

        timeseries = TimeSeriesData(df)
        change_points_vectorized_ = VectorizedCUSUMDetector(timeseries).detector_()

        # take the change points in all columns with the corresponding directions
        change_points_vectorized = [[], []]
        for i in range(len(change_points_vectorized_)):
            for change_points_ts in change_points_vectorized_[i]:
                if change_points_ts.direction == df.columns.values[i]:
                    change_points_vectorized[i].append(change_points_ts)
        # change points for the first column in the matrix
        self.inc_change_points_vectorized = change_points_vectorized[0]
        # change points for the second column in the matrix
        self.dec_change_points_vectorized = change_points_vectorized[1]

        self.dec_change_points_vectorized_int_window = VectorizedCUSUMDetector(timeseries).detector_(
            change_directions=["decrease"], interest_window=(35, 55)
        )[1]

    def test_vectorized_results(self) -> None:
        # check if vectorized CUSUM produces the same results with the original CUSUM
        self.assertEqual(
            self.inc_change_points[0].start_time,
            self.inc_change_points_vectorized[0].start_time,
        )
        self.assertEqual(
            len(self.inc_change_points),
            len(self.inc_change_points_vectorized),
        )
        self.assertEqual(
            self.dec_change_points[0].start_time,
            self.dec_change_points_vectorized[0].start_time,
        )
        self.assertEqual(
            len(self.dec_change_points),
            len(self.dec_change_points_vectorized),
        )
        self.assertEqual(
            self.dec_change_points_int_window[0].start_time,
            self.dec_change_points_vectorized_int_window[0].start_time,
        )
        self.assertEqual(
            len(self.dec_change_points_int_window),
            len(self.dec_change_points_vectorized_int_window),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""This is a file with functions which turn time series into ML features.

Typical use case is to create various features for the nowcasting model.
The features are rolling, i.e. they are the times series as well.

  Typical usage example:

  >>> df = ROC(df, 5)
"""

from typing import Dict
from typing import Optional

import numpy as np
import pandas as pd


def ROC(df: pd.DataFrame, n: int, column: str = "y") -> pd.DataFrame:
    """Adds another column indicating return comparing to step n back.

    Args:
        df: a pandas dataframe.
        n: an integer on how many steps looking back.
        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added column.
    """

    M = df[column].diff(n - 1)
    N = df[column].shift(n - 1)
    if column == "y":
        ROC = pd.Series(M / N, name="ROC_" + str(n), copy=False)
    else:
        ROC = pd.Series(M / N, name=column + "_ROC_" + str(n), copy=False)
    df = df.join(ROC)
    return df


def MOM(df: pd.DataFrame, n: int, column: str = "y") -> pd.DataFrame:
    """Adds another column indicating momentum: difference of current value and n steps back.

    Args:
        df: a pandas dataframe.
        n: an integer on how many steps looking back.
        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added column.
    """

    if column == "y":
        M = pd.Series(df[column].diff(n), name="MOM_" + str(n), copy=False)
    else:
        M = pd.Series(df[column].diff(n), name=column + "_MOM_" + str(n), copy=False)
    df = df.join(M)
    return df


def MA(df: pd.DataFrame, n: int, column: str = "y") -> pd.DataFrame:
    """Adds another column indicating moving average in the past n steps.

    Args:
        df: a pandas dataframe.
        n: an integer on how many steps looking back.
        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added column.
    """

    if column == "y":
        MA = pd.Series(df[column].rolling(n).mean(), name="MA_" + str(n), copy=False)
    else:
        MA = pd.Series(df[column].rolling(n).mean(), name=column + "_MA_" + str(n), copy=False)
    df = df.join(MA)
    return df


def LAG(df: pd.DataFrame, n: int, column: str = "y") -> pd.DataFrame:
    """Adds another column indicating lagged value at the past n steps.

    Args:
        df: a pandas dataframe.
        n: an integer on how many steps looking back.
        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added column.
    """

    N = df[column].shift(n)
    if column == "y":
        LAG = pd.Series(N, name="LAG_" + str(n), copy=False)
    else:
        LAG = pd.Series(N, name=column + "_LAG_" + str(n), copy=False)
    df = df.join(LAG)
    return df


def MACD(df: pd.DataFrame, n_fast: int = 12, n_slow: int = 21, column: str = "y") -> pd.DataFrame:
    """Adds three columns indicating MACD: https://www.investopedia.com/terms/m/macd.asp.

    Args:
        df: a pandas dataframe
        n_fast: an integer on how many steps looking back fast.
        n_slow: an integer on how many steps looking back slow.
        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added 3 columns.
    """

    EMAfast = pd.Series(df[column].ewm(span=n_fast, min_periods=n_slow - 1).mean(), copy=False)
    EMAslow = pd.Series(df[column].ewm(span=n_slow, min_periods=n_slow - 1).mean(), copy=False)
    if column == "y":
        MACD = pd.Series(
            EMAfast - EMAslow,
            name="MACD_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
        MACDsign = pd.Series(
            MACD.ewm(span=9, min_periods=8).mean(),
            name="MACDsign_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
        MACDdiff = pd.Series(
            MACD - MACDsign,
            name="MACDdiff_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
    else:
        MACD = pd.Series(
            EMAfast - EMAslow,
            name=column + "_MACD_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
        MACDsign = pd.Series(
            MACD.ewm(span=9, min_periods=8).mean(),
            name=column + "_MACDsign_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
        MACDdiff = pd.Series(
            MACD - MACDsign,
            name=column + "_MACDdiff_" + str(n_fast) + "_" + str(n_slow),
            copy=False,
        )
    df = df.join(MACD)
    df = df.join(MACDsign)
    df = df.join(MACDdiff)
    return df

    # pyre-fixme[3]: Return type must be annotated.
    def _ewma(arr: np.ndarray, span: int, min_periods: int):
        """
        Exponentialy weighted moving average specified by a decay ``window``
        to provide better adjustments for small windows via:
            y[t] = (x[t] + (1-a)*x[t-1] + (1-a)^2*x[t-2] + ... + (1-a)^n*x[t-n]) /
                   (1 + (1-a) + (1-a)^2 + ... + (1-a)^n).

        Args:
        arr : np.ndarray; A single dimenisional numpy array.
        span : int; The decay window, or 'span'.
        min_periods: int; Minimum amount of data points we'd like to include in the output.

        Returns:
            A np.ndarray. The exponentially weighted moving average of the array.
        """
        output_array = np.empty(arr.shape[0], dtype=np.float64)
        output_array[:] = np.NaN

        arr = arr[~np.isnan(arr)]
        n = arr.shape[0]
        ewma = np.empty(n, dtype=np.float64)
        alpha = 2 / float(span + 1)
        w = 1
        ewma_old = arr[0]
        ewma[0] = ewma_old
        for i in range(1, n):
            w += (1 - alpha) ** i
            ewma_old = ewma_old * (1 - alpha) + arr[i]
            ewma[i] = ewma_old / w

        output_subset = ewma[(min_periods - 1) :]
        output_array[-len(output_subset) :] = output_subset
        return output_array

    # pyre-fixme[3]: Return type must be annotated.
    def _get_nowcasting_np(
        x: np.ndarray,
        window: int = 5,
        n_fast: int = 12,
        n_slow: int = 21,
        extra_args: Optional[Dict[str, bool]] = None,
        default_status: bool = True,
    ):
        """
        Internal function for actually performing feature engineering using the same procedure as
        nowcasting feature_extraction under kats/models.

        Args:
            x: The univariate time series array in the form of 1d numpy array.
            window: int; Length of window size for all Nowcasting features. Default value is 5.
            n_fast: int; length of "fast" or short period exponential moving average in the MACD
                algorithm in the nowcasting features. Default value is 12.
            n_slow: int; length of "slow" or long period exponential moving average in the MACD
                algorithm in the nowcasting features. Default value is 21.
            extra_args: A dictionary containing information for disabling calculation
                of a certain feature. Default value is None, i.e. no feature is disabled.
            default_status: Default status of the switch for calculate the features or not.
                Default value is True.

        Returns:
            A list containing extracted nowcast features.
        """

        # initializing the outputs
        nowcasting_features = [np.nan for _ in range(7)]

        # ROC: indicating return comparing to step n back.
        if extra_args is not None and extra_args.get("nowcast_roc", default_status):
            M = x[(window - 1) :] - x[: -(window - 1)]
            N = x[: -(window - 1)]
            arr = M / N
            nowcasting_features[0] = np.nan_to_num(arr, nan=0.0, posinf=0.0, neginf=0.0).mean()

        # MOM: indicating momentum: difference of current value and n steps back.
        if extra_args is not None and extra_args.get("nowcast_mom", default_status):
            M = x[window:] - x[:-window]
            nowcasting_features[1] = np.nan_to_num(M, nan=0.0, posinf=0.0, neginf=0.0).mean()

        # MA: indicating moving average in the past n steps.
        if extra_args is not None and extra_args.get("nowcast_ma", default_status):
            ret = np.cumsum(x, dtype=float)
            ret[window:] = ret[window:] - ret[:-window]
            ma = ret[window - 1 :] / window
            nowcasting_features[2] = np.nan_to_num(ma, nan=0.0, posinf=0.0, neginf=0.0).mean()

        # LAG: indicating lagged value at the past n steps.
        if extra_args is not None and extra_args.get("nowcast_lag", default_status):
            N = x[:-window]
            nowcasting_features[3] = np.nan_to_num(N, nan=0.0, posinf=0.0, neginf=0.0).mean()

        # MACD: https://www.investopedia.com/terms/m/macd.asp.
        ema_fast = _ewma(x, n_fast, n_slow - 1)
        ema_slow = _ewma(x, n_slow, n_slow - 1)
        MACD = ema_fast - ema_slow
        if extra_args is not None and extra_args.get("nowcast_macd", default_status):
            nowcasting_features[4] = np.nan_to_num(np.nanmean(MACD), nan=0.0, posinf=0.0, neginf=0.0)

        MACDsign = _ewma(MACD, 9, 8)
        if extra_args is not None and extra_args.get("nowcast_macdsign", default_status):
            nowcasting_features[5] = np.nan_to_num(np.nanmean(MACDsign), nan=0.0, posinf=0.0, neginf=0.0)

        MACDdiff = MACD - MACDsign
        if extra_args is not None and extra_args.get("nowcast_macddiff", default_status):
            nowcasting_features[6] = np.nan_to_num(np.nanmean(MACDdiff), nan=0.0, posinf=0.0, neginf=0.0)

        return nowcasting_features


# pyre-fixme[2]: Parameter must be annotated.
def BBANDS(df, n: int, column: str = "y") -> pd.DataFrame:
    """Adds two Bolllinger Band columns

    A Bollinger Band is a technical analysis tool defined by a
    set of trendlines plotted two standard deviations (positively and negatively)
    away from a simple moving average (SMA) of a security's price, but which can
    be adjusted to user preferences.

    Args:
        df: a pandas dataframe
        n: an integer on how many steps looking back.

        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added 2 BollingerBand columns.
    """
    close = df[column]
    MA = pd.Series(close.rolling(n).mean(), copy=False)
    MSD = pd.Series(close.rolling(n).std(), copy=False)
    b1 = 4 * MSD / MA
    B1 = pd.Series(b1, name="BollingerBand1_" + str(n), copy=False)
    df = df.join(B1)
    b2 = (close - MA + 2 * MSD) / (4 * MSD)
    B2 = pd.Series(b2, name="BollingerBand2_" + str(n), copy=False)
    df = df.join(B2)
    return df


# pyre-fixme[2]: Parameter must be annotated.
def TRIX(df, n: int, column: str = "y") -> pd.DataFrame:
    """Adds the TRIX indicator column

    The triple exponential average (TRIX) indicator is an oscillator
    used to identify oversold and overbought markets, and it can
    also be used as a momentum indicator.

    Args:
        df: a pandas dataframe
        n: an integer on how many steps looking back.

        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added TRIX column.
    """
    close = df[column]
    EX1 = close.ewm(span=n, min_periods=n - 1).mean()
    EX2 = EX1.ewm(span=n, min_periods=n - 1).mean()
    EX3 = EX2.ewm(span=n, min_periods=n - 1).mean()

    i = 0
    ROC_l = [0]
    while i + 1 <= df.index[-1]:
        ROC = (EX3[i + 1] - EX3[i]) / EX3[i]
        ROC_l.append(ROC)
        i = i + 1
    Trix = pd.Series(ROC_l, name="TRIX_" + str(n), copy=False)
    df = df.join(Trix)
    return df


# pyre-fixme[2]: Parameter must be annotated.
def EMA(df, n: int, column: str = "y") -> pd.DataFrame:
    """Adds the Exponetial Moving Average column

    The exponential moving average (EMA) is a technical chart indicator that
    tracks the price of an investment (like a stock or commodity) over time.
    The EMA is a type of weighted moving average (WMA) that gives more
     weighting or importance to recent price data

    Args:
        df: a pandas dataframe
        n: an integer on how many steps looking back.

        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added EMA column.
    """
    close = df[column]
    EMA = pd.Series(close.ewm(span=n, min_periods=n - 1).mean(), name="EMA_" + str(n), copy=False)
    df = df.join(EMA)
    return df


# pyre-fixme[2]: Parameter must be annotated.
def TSI(df, r: int, s: int, column: str = "y") -> pd.DataFrame:
    """Adds the TSI column

    The true strength index (TSI) is a technical momentum oscillator
    used to identify trends and reversals

    Args:
        df: a pandas dataframe
        r: an integer on how many steps looking back, for window 1.
        s: an integer on how many steps looking back, for window 2.

        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added TSI column.
    """
    close = df[column]
    M = pd.Series(close.diff(1), copy=False)
    aM = abs(M)
    EMA1 = pd.Series(M.ewm(span=r, min_periods=r - 1).mean(), copy=False)
    aEMA1 = pd.Series(aM.ewm(span=r, min_periods=r - 1).mean(), copy=False)
    EMA2 = pd.Series(EMA1.ewm(span=s, min_periods=s - 1).mean(), copy=False)
    aEMA2 = pd.Series(aEMA1.ewm(span=s, min_periods=s - 1).mean(), copy=False)
    TSI = pd.Series(EMA2 / aEMA2, name="TSI_" + str(r) + "_" + str(s), copy=False)
    df = df.join(TSI)
    return df


# pyre-fixme[2]: Parameter must be annotated.
def RSI(df, n: int, column: str = "y") -> pd.DataFrame:
    """
    Relative Strength Index (RSI)
    Compares the magnitude of recent gains and losses over a specified time
    period to measure speed and change of price movements of a security. It is
    primarily used to attempt to identify overbought or oversold conditions in
    the trading of an asset.

    Args:
        df: a pandas dataframe
        n: an integer on how many steps looking back.

        column: Optional. If column is provided, will calculate based on provided column
            otherwise the column named y will be the target.

    Returns:
        A dataframe with all the columns from input df, and the added RSI column.

    """
    close = df[column]
    diff = close.diff(1)
    up_direction = diff.where(diff > 0, 0.0)
    down_direction = -diff.where(diff < 0, 0.0)
    min_periods = n
    emaup = up_direction.ewm(alpha=1 / n, min_periods=min_periods, adjust=False).mean()
    emadn = down_direction.ewm(alpha=1 / n, min_periods=min_periods, adjust=False).mean()
    relative_strength = emaup / emadn
    rsi_col = pd.Series(
        np.where(emadn == 0, 100, 100 - (100 / (1 + relative_strength))),
        name="RSI_" + str(n),
        copy=False,
    )
    df = df.join(rsi_col)
    return df



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Module with generic outlier detection models. Supports a univariate algorithm that
treates each metric separately to identify outliers and a multivariate detection
algorithm that determines outliers based on joint distribution of metrics
"""

import datetime as dt
import logging
import sys
import traceback
from datetime import datetime
from enum import Enum
from importlib import import_module
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats
from scipy.spatial import distance
from statsmodels.tsa.seasonal import seasonal_decompose

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.consts import TimeSeriesIterator
from kats.detectors.detector import Detector
from kats.models.bayesian_var import BayesianVAR
from kats.models.var import VARModel


class OutlierDetector(Detector):
    """
    This univariate outlier detection algorithm mimics the outlier
    detection algorithm in R

    Attributes:
        data: TimeSeriesData object with the time series
        decomp : 'additive' or 'multiplicative' decomposition
        iqr_mult : iqr_mult * inter quartile range is used to classify outliers
    """

    # pyre-fixme
    outliers: Optional[List[List]] = None
    output_scores: Optional[pd.DataFrame] = None
    decomp: str

    def __init__(self, data: TimeSeriesData, decomp: str = "additive", iqr_mult: float = 3.0) -> None:
        super().__init__(data)
        if decomp in ["additive", "multiplicative"]:
            self.decomp = decomp
        else:
            logging.info("Invalid decomposition setting specified")
            logging.info("Defaulting to Additive Decomposition")
            self.decomp = "additive"
        self.iqr_mult = iqr_mult

    def __clean_ts__(
        self, original: Union[pd.Series, pd.DataFrame]
    ) -> Tuple[List[int], List[float], List[pd.Timestamp]]:
        """
        Performs detection for a single metric. First decomposes the time series
        and detects outliers when the values in residual time series are beyond the
        specified multiplier times the inter quartile range
        Args:
            original: original time series as DataFrame
        Returns: List of detected outlier timepoints in each metric
        """

        original.index = pd.to_datetime(original.index)

        if pd.infer_freq(original.index) is None:
            original = cast(pd.DataFrame, original.asfreq("D"))
            logging.info("Setting frequency to Daily since it cannot be inferred")

        original = cast(
            pd.DataFrame,
            original.interpolate(method="polynomial", limit_direction="both", order=3),
        )

        # This is a hack since polynomial interpolation is not working here
        if sum((np.isnan(x) for x in original["y"])):
            original = cast(
                pd.DataFrame,
                original.interpolate(method="linear", limit_direction="both"),
            )

        # If interpolation added rows, need to return interpolated time index
        output_time = original.index

        # Once our own decomposition is ready, we can directly use it here
        result = seasonal_decompose(original, model=self.decomp, extrapolate_trend="freq")
        rem = result.resid
        detrend = original["y"] - result.trend
        strength = float(1 - np.nanvar(rem) / np.nanvar(detrend))
        if strength >= 0.6:
            original["y"] = original["y"] - result.seasonal
        # using IQR as threshold
        resid = original["y"] - result.trend
        resid_q = np.nanpercentile(resid, [25, 75])
        iqr = resid_q[1] - resid_q[0]
        limits = resid_q + (self.iqr_mult * iqr * np.array([-1, 1]))
        # calculate scores
        output_scores = list((resid - limits[0]) / (limits[1] - limits[0]))
        outliers = resid[(resid > limits[1]) | (resid < limits[0])]
        outliers_index = list(outliers.index)

        return outliers_index, output_scores, output_time

    # pyre-fixme
    def detector(self, method: Optional[str] = None) -> None:
        """
        Detects outliers and stores in self.outliers
        """

        self.iter = ts_iter = TimeSeriesIterator(self.data)
        column_names = [col for col in self.data.to_dataframe().columns if col != self.data.time_col_name]
        self.outliers = outliers = []
        output_scores_dict = {}
        time_index = None
        logging.Logger("Detecting Outliers")

        for ts, col in zip(ts_iter, column_names):
            try:
                outliers_index, output_scores, time_index = self.__clean_ts__(ts)
                outliers.append(outliers_index)
                output_scores_dict[col] = output_scores
            except BaseException:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                logging.error("".join("!! " + line for line in lines))
                logging.error("Outlier Detection Failed")
                outliers.append([])

        if output_scores_dict:
            self.output_scores = pd.DataFrame(output_scores_dict, index=time_index, copy=False)


class MultivariateAnomalyDetectorType(Enum):
    VAR = "var.VARModel"
    BAYESIAN_VAR = "bayesian_var.BayesianVAR"


class MultivariateAnomalyDetector(Detector):
    """
    Detector class for Multivariate Outlier Detection.
    Provides utilities to calculate anomaly scores, get anomalous
    datapoints and anomalous metrics at those points.

    Attributes:
        data: Input metrics TimeSeriesData
        params: Parameter class for multivariate VAR/ BVAR model
        training_days: num of days of data to use for initial training.
                    If less than a day, specify as fraction of a day
        model_type: The type of multivariate anomaly detector (currently 'BAYESIAN_VAR' and 'VAR' options available)
    """

    resid: Optional[np.ndarray] = None
    sigma_u: Optional[np.ndarray] = None
    anomaly_score_df: Optional[pd.DataFrame] = None

    def __init__(
        self,
        data: TimeSeriesData,
        params: Params,
        training_days: float,
        model_type: MultivariateAnomalyDetectorType = MultivariateAnomalyDetectorType.VAR,
    ) -> None:
        super().__init__(data)
        df = data.to_dataframe().set_index("time")
        self.df: pd.DataFrame = df

        params.validate_params()
        self.params = params

        # pyre-fixme
        time_diff = data.time.sort_values().diff().dropna()
        if len(time_diff.unique()) == 1:  # check constant frequenccy
            freq = time_diff.unique()[0].astype("int")
            self.granularity_days: float = freq / (24 * 3600 * (10**9))
        else:
            raise RuntimeError("Frequency of metrics is not constant." "Please check for missing or duplicate values")

        self.training_days = training_days
        self.detector_model = model_type

    def _clean_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Remove outliers in training data based on zscore

        Args:
            df: input dataframe

        Returns: Clean dataframe
        """

        z_score_threshold = 3
        # pyre-fixme
        zscore_df = stats.zscore(df)
        non_outlier_flag = zscore_df < z_score_threshold
        df_clean = df.where(non_outlier_flag, np.nan)
        df_clean = df_clean.interpolate(method="linear", order=2, limit_direction="both")
        return df_clean

    def _is_pos_def(self, mat: np.ndarray) -> bool:
        """
        Check if matrix is positive definite.

        Args:
            mat: numpy matrix

        Returns:
            True if mat is positive definite
        """
        return np.all(np.linalg.eigvals(mat) > 0)

    def _create_model(self, data: TimeSeriesData, params: Params) -> Union[VARModel, BayesianVAR]:
        model_name = f"kats.models.{self.detector_model.value}"
        module_name, model_name = model_name.rsplit(".", 1)
        return getattr(import_module(module_name), model_name)(data, params)

    def _generate_forecast(self, t: datetime) -> pd.DataFrame:
        """
        Fit VAR model and generate 1 step ahead forecasts (t+1)

        Args:
            t: time until which to use for training the VAR model

        Returns:
            DataFrame with predicted expected value for each metric value at (t+1)
        """
        logging.info(f"Generating forecasts at {t}")
        look_back = dt.timedelta(days=self.training_days)
        train = self.df.loc[t - look_back : t, :]
        train_clean = self._clean_data(train)

        # fit VAR
        model = self._create_model(TimeSeriesData(train_clean.reset_index()), self.params)
        model.fit()
        lag_order = model.k_ar
        logging.info(f"Fitted VAR model of order {lag_order}")
        # model.resid and model.sigma_u set in model.fit(), cast from Optional[np.ndarray] to np.ndarray
        self.resid = cast(np.ndarray, model.resid)
        self.sigma_u = sigma_u = cast(np.ndarray, model.sigma_u)
        if ~(self._is_pos_def(sigma_u)):
            msg = f"Fitted Covariance matrix at time {t} is not positive definite"
            logging.error(msg)
            raise RuntimeError(msg)

        # predict
        pred = model.predict(steps=1)
        forecast = [[k, float(pred[k].value["fcst"])] for k, v in pred.items()]
        pred_df = pd.DataFrame(columns=["index", "est"], data=forecast, copy=False).set_index("index")
        test = self.df.loc[t + dt.timedelta(days=self.granularity_days), :]
        pred_df["actual"] = test

        return pred_df

    def _calc_anomaly_scores(self, pred_df: pd.DataFrame) -> Dict[str, float]:
        """
        Calculate overall anomay score at time t based on multivariate Mahalonabis distance
        and individual anomaly scores as zscores

        Args:
            pred_df: Dataframe with forecasted values

        Returns:
            Dictionary with overall and individual anomaly scores along with p-value
        """

        # individual anomaly scores
        cov = self.sigma_u
        resid = self.resid
        assert cov is not None and resid is not None
        residual_score = {}
        rt = pred_df["est"] - pred_df["actual"]
        for col in cov.columns:
            residual_mean = resid[col].mean()
            residual_var = resid[col].var()
            residual_score[col] = np.abs((rt[col] - residual_mean)) / np.sqrt(residual_var)

        # overall anomaly score
        cov_inv = np.linalg.inv(cov.values)
        overall_anomaly_score = distance.mahalanobis(rt.values, resid.mean().values, cov_inv)
        residual_score["overall_anomaly_score"] = overall_anomaly_score
        # calculate p-values
        dof = len(self.df.columns)
        # pyre-fixme
        residual_score["p_value"] = stats.chi2.sf(overall_anomaly_score, df=dof)

        return residual_score

    # pyre-fixme
    def detector(self) -> pd.DataFrame:
        """
        Fit the detection model and return the results

        Returns:
            DataFrame with colums corresponding to individual anomaly scores
            of each metric and the overall anomaly score for the whole timeseries
        """
        anomaly_score_df = pd.DataFrame()
        look_back = dt.timedelta(days=self.training_days)
        fcstTime = self.df.index.min() + look_back
        while fcstTime < self.df.index.max():
            # forecast for fcstTime+ 1
            pred_df = self._generate_forecast(fcstTime)
            # calculate anomaly scores
            anomaly_scores_t = self._calc_anomaly_scores(pred_df)
            # process next observation
            fcstTime += dt.timedelta(days=self.granularity_days)
            anomaly_scores_t = pd.DataFrame(anomaly_scores_t, index=[fcstTime], copy=False)
            anomaly_score_df = anomaly_score_df.append(anomaly_scores_t)

        self.anomaly_score_df = anomaly_score_df
        return anomaly_score_df

    def get_anomaly_timepoints(self, alpha: float) -> List[pd.DatetimeIndex]:
        """
        Helper function to get anomaly timepoints based on the significance level

        Args:
            alpha: significance level to consider the timeperiod anomalous
            Use .plot() to help choose a good threshold

        Returns:
            List of time instants when the system of metrics show anomalous behavior
        """
        anomaly_score_df = self.anomaly_score_df
        if anomaly_score_df is None:
            raise ValueError("detector() must be called before get_anomaly_timepoints()")

        flag = anomaly_score_df.p_value < alpha
        anomaly_ts = anomaly_score_df[flag].index

        return list(anomaly_ts)

    def get_anomalous_metrics(self, t: datetime, top_k: Optional[int] = None) -> pd.DataFrame:
        """
        Find top k metrics with most anomalous behavior at time t

        Args:
            t: time instant of interest (same type as TimeSeriesData.time)
            top_k: Top few metrics to return. If None, returns all

        Returns:
            DataFrame with metrics and their corresponding anomaly score
        """
        anomaly_score_df = self.anomaly_score_df
        if anomaly_score_df is None:
            raise ValueError("detector() must be called before get_anomalous_metrics()")

        residual_scores = anomaly_score_df.drop(columns="overall_anomaly_score")
        residual_scores_t = residual_scores.loc[t, :].sort_values(ascending=False).reset_index()
        residual_scores_t.columns = ["metric", "anomaly_score"]

        return residual_scores_t[:top_k]

    def plot(self, figsize: Optional[Tuple[int, int]] = None, **kwargs: Any) -> Tuple[plt.Axes, plt.Axes]:
        """
        Plot overall anomaly score of system of metrics at each instant.
        Useful for threshold selection

        Args:
            figsize: figure size. If None, use (15, 16).
        Returns:
            The matplotlib Axes.
        """
        anomaly_score_df = self.anomaly_score_df
        if anomaly_score_df is None:
            raise ValueError("detector() must be called before get_anomalous_metrics()")

        if figsize is None:
            figsize = (15, 8 * 2)
        _, axs = plt.subplots(2, 1, sharex=True, figsize=figsize)
        a = pd.merge(
            self.df,
            anomaly_score_df["overall_anomaly_score"],
            left_index=True,
            right_index=True,
            how="right",
            copy=False,
        )
        axs[0].plot(a.drop(columns=["overall_anomaly_score"]))
        axs[0].set_title("Input time series metrics")

        axs[1].plot(a["overall_anomaly_score"])
        axs[1].set_title("Overall Anomaly Score")
        return axs



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Defines the base class for detectors.
"""

from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.consts import TimeSeriesIterator
from kats.detectors.detector_consts import AnomalyResponse


class Detector(ABC):
    """Base detector class to be inherited by specific detectors

    Attributes:
        data: The input time series data from TimeSeriesData
    """

    iter: Optional[TimeSeriesIterator] = None
    outliers: Optional[Sequence[TimeSeriesChangePoint]] = None

    def __init__(self, data: TimeSeriesData) -> None:
        self.data = data
        self.__type__ = "detector"
        if data is not None:
            self.data.time = pd.to_datetime(self.data.time)

    @abstractmethod
    def detector(self, method: Optional[str] = None) -> Sequence[TimeSeriesChangePoint]:
        raise NotImplementedError()

    def remover(self, interpolate: bool = False) -> TimeSeriesData:
        """Remove the outlier in time series

        Args:
            interpolate: Optional; bool; interpolate the outlier

        Returns:
            A TimeSeriesData with outlier removed.
        """
        outliers = self.outliers
        if outliers is None:
            return self.data

        df = []
        self.detector()
        count = 0
        for count, ts in enumerate(TimeSeriesIterator(self.data)):
            ts.loc[outliers[count], "y"] = np.nan
            df.append(ts)

        # Need to make this a ts object
        df_final = pd.concat(df, axis=1, copy=False)

        if interpolate:
            df_final.interpolate(method="linear", limit_direction="both", inplace=True)
        # may contain multiple time series y
        df_final.columns = [f"y_{i}" for i in range(count + 1)]
        df_final["time"] = df_final.index
        ts_out = TimeSeriesData(df_final)
        return ts_out

    def plot(self, **kwargs: Any) -> Union[plt.Axes, Sequence[plt.Axes]]:
        raise NotImplementedError()


class DetectorModel(ABC):
    """
    Base Detector model class to be inherited by specific detectors. A DetectorModel
    keeps the state of the Detector, and implements the incremental model training.

    The usage of the DetectorModel is (replace DetectorModel with the proper child class)

    model = DetectorModel(serialized_model)
    model.fit(new_data, ...)

    # the model may be saved through model.serialize() call
    # the model may be loaded again through model = DetectorModel(serialized_model)

    result = model.predict(data, ...)
    """

    @abstractmethod
    def __init__(self, serialized_model: Optional[bytes]) -> None:
        raise NotImplementedError()

    """
    Serialize the model. It's required that the serialized model can be unserialized
    by the next version of the same DetectorModel class. During upgrade of a model class,
    version 1 to 2, the version 2 code will unserialize version 1 model, create the
    new model (version 2) instance, the serialize out the version 2 instance, thus
    completing the upgrade.
    """

    @abstractmethod
    def serialize(self) -> bytes:
        raise NotImplementedError()

    """
    Fit the model with the data passes in and update the model's state.
    """

    @abstractmethod
    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData],
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError()

    """
    Given the time series data, returns the anomaly score time series data with
    matching timestamps.
    """

    @abstractmethod
    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData],
        **kwargs: Any,
    ) -> AnomalyResponse:
        raise NotImplementedError()

    """
    This method will change the state and return the anomaly scores.
    """

    @abstractmethod
    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData],
        **kwargs: Any,
    ) -> AnomalyResponse:
        raise NotImplementedError()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from operator import attrgetter
from unittest import TestCase

import pandas as pd
from parameterized.parameterized import parameterized

from kats.detectors.trend_mk import MKDetector
from kats.tests.detectors.utils import gen_no_trend_data_ndim
from kats.tests.detectors.utils import gen_trend_data_ndim


class UnivariateMKDetectorTest(TestCase):
    def setUp(self) -> None:
        self.window_size = 20
        self.time = pd.Series(pd.date_range(start="2020-01-01", end="2020-06-20", freq="1D"))
        no_trend_data = gen_no_trend_data_ndim(time=self.time)
        trend_data, self.t_change = gen_trend_data_ndim(time=self.time)
        trend_seas_data, self.t_change_seas = gen_trend_data_ndim(time=self.time, seasonality=0.07)

        # no trend data
        self.d_no_trend = MKDetector(data=no_trend_data)
        self.detected_time_points_no_trend = self.d_no_trend.detector(window_size=self.window_size)

        # trend data
        self.d_trend = MKDetector(data=trend_data)
        self.detected_time_points_trend = self.d_trend.detector(window_size=self.window_size)
        self.metadata_trend = self.detected_time_points_trend[0]
        results_trend = self.d_trend.get_MK_statistics()
        self.up_trend_detected_trend = self.d_trend.get_MK_results(results_trend, direction="up")["ds"]
        self.down_trend_detected_trend = self.d_trend.get_MK_results(results_trend, direction="down")["ds"]

        # trend data anchor point
        self.detected_time_points_trend2 = self.d_trend.detector(training_days=30)
        results_trend2 = self.d_trend.get_MK_statistics()
        self.up_trend_detected_trend2 = self.d_trend.get_MK_results(results_trend2, direction="up")["ds"]
        self.down_trend_detected_trend2 = self.d_trend.get_MK_results(results_trend2, direction="down")["ds"]

        # trend data with seasonality
        self.d_seas = MKDetector(data=trend_seas_data)
        self.detected_time_points_seas = self.d_seas.detector(freq="weekly")
        results_seas = self.d_seas.get_MK_statistics()
        self.up_trend_detected_seas = self.d_seas.get_MK_results(results_seas, direction="up")["ds"]
        self.down_trend_detected_seas = self.d_seas.get_MK_results(results_seas, direction="down")["ds"]

        # trend data with seasonality anchor point
        self.detected_time_points_seas2 = self.d_seas.detector(training_days=30, freq="weekly")
        results_seas2 = self.d_seas.get_MK_statistics()
        self.up_trend_detected_seas2 = self.d_seas.get_MK_results(results_seas2, direction="up")["ds"]
        self.down_trend_detected_seas2 = self.d_seas.get_MK_results(results_seas2, direction="down")["ds"]

    # test for no trend data
    def test_no_trend_data(self) -> None:
        self.assertEqual(len(self.detected_time_points_no_trend), 0)

    # test for trend data
    def test_detector_type(self) -> None:
        self.assertIsInstance(self.d_trend, self.metadata_trend.detector_type)

    def test_tau(self) -> None:
        self.assertIsInstance(self.metadata_trend.Tau, float)

    def test_is_univariate(self) -> None:
        self.assertFalse(self.metadata_trend.is_multivariate)

    def test_incr_trend(self) -> None:
        self.assertEqual(self.metadata_trend.trend_direction, "increasing")

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["up_trend_detected_trend"],
    #  ["up_trend_detected_seas"]])`.
    @parameterized.expand([["up_trend_detected_trend"], ["up_trend_detected_seas"]])
    def test_upward_after_start(self, up_trend_detected: str) -> None:
        self.assertGreaterEqual(
            attrgetter(up_trend_detected)(self).iloc[0],
            self.time[0],
            msg=f"The first {self.window_size}-days upward trend was not detected after it starts.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["up_trend_detected_trend",
    #  "t_change"], ["up_trend_detected_seas", "t_change_seas"]])`.
    @parameterized.expand(
        [
            ["up_trend_detected_trend", "t_change"],
            ["up_trend_detected_seas", "t_change_seas"],
        ]
    )
    def test_upward_before_end(self, up_trend_detected: str, t_change: str) -> None:
        self.assertLessEqual(
            attrgetter(up_trend_detected)(self).iloc[-1],
            self.time[attrgetter(t_change)(self)[0] + self.window_size],
            msg=f"The last {self.window_size}-days upward trend was not detected before it ends.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["down_trend_detected_trend",
    #  "t_change"], ["down_trend_detected_seas", "t_change_seas"]])`.
    @parameterized.expand(
        [
            ["down_trend_detected_trend", "t_change"],
            ["down_trend_detected_seas", "t_change_seas"],
        ]
    )
    def test_downward_after_start(self, down_trend_detected: str, t_change: str) -> None:
        self.assertGreaterEqual(
            attrgetter(down_trend_detected)(self).iloc[0],
            self.time[attrgetter(t_change)(self)[0]],
            msg=f"The first {self.window_size}-days downward trend was not detected after it starts.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["down_trend_detected_trend"],
    #  ["down_trend_detected_trend2"], ["down_trend_detected_seas"],
    #  ["down_trend_detected_seas2"]])`.
    @parameterized.expand(
        [
            ["down_trend_detected_trend"],
            ["down_trend_detected_trend2"],
            ["down_trend_detected_seas"],
            ["down_trend_detected_seas2"],
        ]
    )
    def test_downward_before_end(self, down_trend_detected: str) -> None:
        self.assertEqual(
            attrgetter(down_trend_detected)(self).iloc[-1],
            self.time[len(self.time) - 1],
            msg=f"The last {self.window_size}-days downward trend was not detected before it ends.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["d_no_trend", "detected_time_points_no_trend"],
            ["d_trend", "detected_time_points_trend"],
            ["d_trend", "detected_time_points_trend2"],
            ["d_seas", "detected_time_points_seas"],
            ["d_seas", "detected_time_points_seas2"],
        ]
    )
    def test_plot(self, detector: str, detected_time_points: str) -> None:
        attrgetter(detector)(self).plot(attrgetter(detected_time_points)(self))


class MultivariateMKDetectorTest(TestCase):
    def setUp(self) -> None:
        self.window_size = 20
        self.time = pd.Series(pd.date_range(start="2020-01-01", end="2020-06-20", freq="1D"))
        self.ndim = 5
        no_trend_data = gen_no_trend_data_ndim(time=self.time, ndim=self.ndim)
        trend_data, self.t_change = gen_trend_data_ndim(time=self.time, ndim=self.ndim)
        trend_seas_data, self.t_change_seas = gen_trend_data_ndim(time=self.time, seasonality=0.07, ndim=self.ndim)

        # no trend data
        self.d_no_trend = MKDetector(data=no_trend_data)
        self.detected_time_points_no_trend = self.d_no_trend.detector(window_size=self.window_size)

        # trend data
        self.d_trend = MKDetector(data=trend_data, multivariate=True)
        self.detected_time_points_trend = self.d_trend.detector()
        results_trend = self.d_trend.get_MK_statistics()
        self.up_trend_detected_trend = self.d_trend.get_MK_results(results_trend, direction="up")["ds"]
        self.down_trend_detected_trend = self.d_trend.get_MK_results(results_trend, direction="down")["ds"]

        # trend data with seasonality
        self.d_seas = MKDetector(data=trend_seas_data, multivariate=True)
        self.detected_time_points_seas = self.d_seas.detector(freq="weekly")
        results_seas = self.d_seas.get_MK_statistics()
        self.up_trend_detected_seas = self.d_seas.get_MK_results(results_seas, direction="up")["ds"]
        self.down_trend_detected_seas = self.d_seas.get_MK_results(results_seas, direction="down")["ds"]

    # test for no trend data
    def test_no_trend_data(self) -> None:
        self.assertEqual(len(self.detected_time_points_no_trend), 0)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["up_trend_detected_trend"],
    #  ["up_trend_detected_seas"]])`.
    @parameterized.expand([["up_trend_detected_trend"], ["up_trend_detected_seas"]])
    def test_upward_after_start(self, up_trend_detected: str) -> None:
        self.assertGreaterEqual(
            attrgetter(up_trend_detected)(self).iloc[0],
            self.time[0],
            msg=f"The first {self.window_size}-days upward trend was not detected after it starts.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["up_trend_detected_trend",
    #  "t_change"], ["up_trend_detected_seas", "t_change_seas"]])`.
    @parameterized.expand(
        [
            ["up_trend_detected_trend", "t_change"],
            ["up_trend_detected_seas", "t_change_seas"],
        ]
    )
    def test_upward_before_end(self, up_trend_detected: str, t_change: str) -> None:
        self.assertLessEqual(
            attrgetter(up_trend_detected)(self).iloc[-1],
            self.time[attrgetter(t_change)(self)[0] + self.window_size],
            msg=f"The last {self.window_size}-days upward trend was not detected before it ends.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["down_trend_detected_trend",
    #  "t_change"], ["down_trend_detected_seas", "t_change_seas"]])`.
    @parameterized.expand(
        [
            ["down_trend_detected_trend", "t_change"],
            ["down_trend_detected_seas", "t_change_seas"],
        ]
    )
    def test_downward_after_start(self, down_trend_detected: str, t_change: str) -> None:
        self.assertGreaterEqual(
            attrgetter(down_trend_detected)(self).iloc[0],
            self.time[attrgetter(t_change)(self)[0]],
            msg=f"The first {self.window_size}-days downward trend was not detected after it starts.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["down_trend_detected_trend"],
    #  ["down_trend_detected_seas"]])`.
    @parameterized.expand([["down_trend_detected_trend"], ["down_trend_detected_seas"]])
    def test_downward_before_end(self, down_trend_detected: str) -> None:
        self.assertEqual(
            attrgetter(down_trend_detected)(self).iloc[-1],
            self.time[len(self.time) - 1],
            msg=f"The last {self.window_size}-days downward trend was not detected before it ends.",
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["d_no_trend",
    #  "detected_time_points_no_trend"], ["d_trend", "detected_time_points_trend"],
    #  ["d_seas", "detected_time_points_seas"]])`.
    @parameterized.expand(
        [
            ["d_no_trend", "detected_time_points_no_trend"],
            ["d_trend", "detected_time_points_trend"],
            ["d_seas", "detected_time_points_seas"],
        ]
    )
    def test_plot(self, detector: str, detected_time_points: str) -> None:
        attrgetter(detector)(self).plot(attrgetter(detected_time_points)(self))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from statsmodels.tsa.seasonal import STL
from statsmodels.tsa.seasonal import seasonal_decompose

from kats.consts import TimeSeriesData

# from numpy.typing import ArrayLike
ArrayLike = Union[np.ndarray, Sequence[float]]
Figsize = Tuple[int, int]


def _identity(x: ArrayLike) -> ArrayLike:
    return x


class TimeSeriesDecomposition:
    """Model class for Time Series Decomposition.

    This class provides utilities to decompose an input time series

    Pass specific arguments to seasonal_decompose and STL functions via kwargs.

    Attributes:
        data: the input time series data as `TimeSeriesData`
        decomposition: `additive` or `multiplicative` decomposition
        method: `STL decompostion` or `seasonal_decompose`
    """

    freq: Optional[Union[str, pd.Timedelta]] = None
    results: Optional[Dict[str, TimeSeriesData]] = None
    decomposition: str
    method: Callable[[pd.DataFrame], Dict[str, pd.DataFrame]]
    period: Optional[int]
    seasonal: int
    trend: Optional[int]
    low_pass: Optional[int]
    seasonal_deg: int
    trend_deg: int
    low_pass_deg: int
    robust: bool
    seasonal_jump: int
    trend_jump: int
    low_pass_jump: int

    def __init__(
        self,
        data: TimeSeriesData,
        decomposition: str = "additive",
        method: str = "STL",
        **kwargs: Any,
    ) -> None:
        if not isinstance(data.value, pd.Series):
            msg = f"Only support univariate time series, but got {type(data.value)}."
            logging.error(msg)
            raise ValueError(msg)
        self.data = data
        if decomposition in ("additive", "multiplicative"):
            self.decomposition = decomposition
        else:
            logging.info("Invalid decomposition setting specified; " "defaulting to Additive Decomposition.")
            self.decomposition = "additive"
        if method == "seasonal_decompose":
            self.method = self.__decompose_seasonal
        else:
            if method != "STL":
                logging.info(
                    f"""Invalid decomposition setting {method} specified.
                    Possible Values: STL, seasonal_decompose.
                    Defaulting to STL."""
                )
            self.method = self.__decompose_STL

        ## The following are params for the STL Module
        self.period = kwargs.get("period", None)
        self.seasonal = kwargs.get("seasonal", 7)
        self.trend = kwargs.get("trend", None)
        self.low_pass = kwargs.get("low_pass", None)
        self.seasonal_deg = kwargs.get("seasonal_deg", 1)
        self.trend_deg = kwargs.get("trend_deg", 1)
        self.low_pass_deg = kwargs.get("low_pass_deg", 1)
        self.robust = kwargs.get("robust", False)
        self.seasonal_jump = kwargs.get("seasonal_jump", 1)
        self.trend_jump = kwargs.get("trend_jump", 1)
        self.low_pass_jump = kwargs.get("low_pass_jump", 1)

    def __clean_ts(self) -> pd.DataFrame:
        """Internal function to clean the time series.

        Internal function to interpolate time series and infer frequency of
        time series required for decomposition.
        """

        original = pd.DataFrame(
            list(self.data.value),
            index=pd.to_datetime(self.data.time),
            columns=["y"],
            copy=False,
        )

        if self.data.infer_freq_robust() is None:
            original = original.asfreq("D")
            logging.info("Setting frequency to Daily since it cannot be inferred")
            self.freq = pd.infer_freq(original.index)
        else:
            self.freq = self.data.infer_freq_robust()

        original.interpolate(method="polynomial", limit_direction="both", order=3, inplace=True)

        ## This is a hack since polynomial interpolation is not working here
        if any(original["y"].isna()):
            original.interpolate(method="linear", limit_direction="both", inplace=True)

        # pyre-ignore[7]: Expected `DataFrame` but got
        #  `Union[pd.core.frame.DataFrame, pd.core.series.Series]`.
        return original

    def _get_period(self) -> Optional[int]:
        period = self.period
        freq = self.freq
        if period is None:
            if freq is not None and isinstance(freq, str) and "T" in freq:
                logging.warning(
                    """Seasonal Decompose cannot handle sub day level granularity.
                    Please consider setting period yourself based on the input data.
                    Defaulting to a period of 2."""
                )
                period = 2
            elif freq is not None and isinstance(freq, pd.Timedelta) and freq.days == 0:
                logging.warning(
                    """Seasonal Decompose cannot handle sub day level granularity.
                    Please consider setting period yourself based on the input data.
                    Defaulting to a period of 2."""
                )
                period = 2
        return period

    def __decompose_seasonal(self, original: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """Internal function to call seasonal_decompose to do the decomposition."""
        period = self._get_period()
        result = seasonal_decompose(original, model=self.decomposition, period=period)

        return {
            "trend": result.trend,
            "seasonal": result.seasonal,
            "rem": result.resid,
        }

    def __decompose_STL(self, original: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """Internal function to call STL to do the decomposition.

        The arguments to STL can be passed in the class via kwargs
        """
        self.period = period = self._get_period()

        if self.decomposition == "additive":
            data = original
            post_transform = _identity
        else:
            if np.any(original <= 0):
                logging.error("Multiplicative seasonality is not appropriate " "for zero and negative values.")
            data = np.log(original)
            post_transform = np.exp

        result = STL(
            data,
            period=period,
            seasonal=self.seasonal,
            trend=self.trend,
            low_pass=self.low_pass,
            seasonal_deg=self.seasonal_deg,
            trend_deg=self.trend_deg,
            low_pass_deg=self.low_pass_deg,
            robust=self.robust,
            seasonal_jump=self.seasonal_jump,
            trend_jump=self.trend_jump,
            low_pass_jump=self.low_pass_jump,
        ).fit()

        return {
            "trend": post_transform(result.trend),
            "seasonal": post_transform(result.seasonal),
            "rem": post_transform(result.resid),
        }

    def __decompose(self, original: pd.DataFrame) -> Dict[str, TimeSeriesData]:
        output = self.method(original)
        return {
            name: TimeSeriesData(ts.reset_index(), time_col_name=self.data.time_col_name)
            for name, ts in output.items()
        }

    def decomposer(self) -> Dict[str, TimeSeriesData]:
        """Decompose the time series.

        Returns:
            A dictionary with three time series for the three components:
            `trend` : Trend
            `seasonal` : Seasonality, and
            `rem` : Residual
        """
        original = self.__clean_ts()
        self.results = result = self.__decompose(original)
        return result

    def plot(
        self,
        figsize: Optional[Figsize] = None,
        linewidth: int = 3,
        xlabel: str = "Time",
        original_title: str = "Original Time Series",
        trend_title: str = "Trend",
        seasonality_title: str = "Seasonality",
        residual_title: str = "Residual",
        subplot_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Tuple[plt.Axes, plt.Axes, plt.Axes, plt.Axes]:
        """Plot the original time series and the three decomposed components."""
        results = self.results
        if results is None:
            raise ValueError("Call decomposer() before plot().")

        if figsize is None:
            figsize = (20, 10)
        if subplot_kwargs is None:
            subplot_kwargs = {"hspace": 0.2}

        sharex = kwargs.pop("sharex", True)
        fig, axs = plt.subplots(nrows=4, ncols=1, figsize=figsize, sharex=sharex, **kwargs)

        axs[0].plot(
            self.data.time.values,
            self.data.value.values,
            linewidth=linewidth,
        )
        axs[0].set_title(original_title)

        titles = [trend_title, seasonality_title, residual_title]
        parts = ["trend", "seasonal", "rem"]
        for part, ax, title in zip(parts, axs[1:], titles):
            ts: TimeSeriesData = results[part]
            ax.plot(ts.time.values, ts.value.values, linewidth=linewidth)
            ax.set_title(title)

        axs[3].set_xlabel(xlabel)
        plt.subplots_adjust(**subplot_kwargs)
        return (axs[0], axs[1], axs[2], axs[3])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import attr
import numpy as np
import pandas as pd
from scipy.stats import norm  # @manual
from scipy.stats import t
from scipy.stats import ttest_ind
from statsmodels.stats import multitest

from kats.consts import TimeSeriesData

# from np.typing import ArrayLike
ArrayLike = np.ndarray


# Single Spike object
@attr.s(auto_attribs=True)
class SingleSpike:
    time: datetime
    value: float
    n_sigma: float

    @property
    def time_str(self) -> str:
        return datetime.strftime(self.time, "%Y-%m-%d")


# Changepoint Interval object
@attr.s(auto_attribs=True)
class ChangePointInterval:
    start_time: datetime
    end_time: datetime
    previous_interval: Optional[ChangePointInterval] = attr.ib(default=None, init=False)
    _all_spikes: Union[Optional[List[SingleSpike]], Optional[List[List[SingleSpike]]]] = attr.ib(
        default=None, init=False
    )
    spike_std_threshold: float = attr.ib(default=2.0, init=False)
    data_df: Optional[pd.DataFrame] = attr.ib(None, init=False)
    _ts_cols: List[str] = attr.ib(factory=lambda: ["value"], init=False)
    num_series: int = 1

    @property
    def data(self) -> Optional[ArrayLike]:
        df = self.data_df
        if df is None:
            return None
        elif self.num_series == 1:
            return df.value.values
        else:
            return df[self._ts_cols].values

    @data.setter
    def data(self, data: TimeSeriesData) -> None:
        if not data.is_univariate():
            self._ts_cols = list(data.value.columns)
            self.num_series = len(self._ts_cols)
        all_data_df = data.to_dataframe()
        all_data_df.columns = ["time"] + self._ts_cols
        all_data_df["time"] = pd.to_datetime(all_data_df["time"])
        all_data_df = all_data_df.loc[(all_data_df.time >= self.start_time) & (all_data_df.time < self.end_time)]
        self.data_df = all_data_df

    def _detect_spikes(self) -> Union[List[SingleSpike], List[List[SingleSpike]]]:
        df = self.data_df
        if df is None:
            raise ValueError("data must be set before spike detection")

        if self.num_series == 1:
            df["z_score"] = (df.value - self.mean_val) / np.sqrt(self.variance_val)

            spike_df = df.query(f"z_score >={self.spike_std_threshold}")
            return [
                SingleSpike(time=row["time"], value=row["value"], n_sigma=row["z_score"])
                for counter, row in spike_df.iterrows()
            ]
        else:
            spikes = []
            for i, c in enumerate(self._ts_cols):
                mean_val, variance_val = self.mean_val, self.variance_val
                if isinstance(mean_val, float) or isinstance(variance_val, float):
                    raise ValueError(
                        f"num_series = {self.num_series} so mean_val and variance_val should have type ArrayLike."
                    )
                df[f"z_score_{c}"] = (df[c] - mean_val[i]) / np.sqrt(variance_val[i])

                spike_df = df.query(f"z_score_{c} >={self.spike_std_threshold}")

                if spike_df.shape[0] == 0:
                    continue
                else:
                    spikes.append(
                        [
                            SingleSpike(
                                time=row["time"],
                                value=row[c],
                                n_sigma=row[f"z_score_{c}"],
                            )
                            for counter, row in spike_df.iterrows()
                        ]
                    )
            return spikes

    def extend_data(self, data: TimeSeriesData) -> None:
        """
        extends the data.
        """
        new_data_df = data.to_dataframe()
        new_data_df.columns = ["time"] + self._ts_cols
        df = self.data_df
        if df is not None:
            new_data_df = pd.concat([df, new_data_df], copy=False)
        self.data_df = new_data_df.loc[(new_data_df.time >= self.start_time) & (new_data_df.time < self.end_time)]

    @property
    def start_time_str(self) -> str:
        return datetime.strftime(self.start_time, "%Y-%m-%d")

    @property
    def end_time_str(self) -> str:
        return datetime.strftime(self.end_time, "%Y-%m-%d")

    @property
    def mean_val(self) -> Union[float, ArrayLike]:
        if self.num_series == 1:
            vals = self.data
            return 0.0 if vals is None else np.mean(vals)
        else:
            data_df = self.data_df
            if data_df is None:
                return np.zeros(self.num_series)
            return np.array([np.mean(data_df[c].values) for c in self._ts_cols])

    @property
    def variance_val(self) -> Union[float, ArrayLike]:
        if self.num_series == 1:
            vals = self.data
            # the t-test uses the sample standard deviation^2 instead of variance,
            return 0.0 if vals is None or len(vals) == 1 else np.var(vals, ddof=1)
        else:
            data_df = self.data_df
            if data_df is None or len(data_df) == 1:
                return np.zeros(self.num_series)
            # the t-test uses the sample standard deviation^2 instead of variance,
            return np.array([np.var(data_df[c].values, ddof=1) for c in self._ts_cols])

    def __len__(self) -> int:
        df = self.data_df
        return 0 if df is None else len(df)

    @property
    def spikes(self) -> Union[List[SingleSpike], List[List[SingleSpike]]]:
        spikes = self._all_spikes
        if spikes is None:
            spikes = self._detect_spikes()
            self._all_spikes = spikes
        return spikes


# Percentage Change Object
class PercentageChange:
    """
    PercentageChange is a class which is widely used in detector models. It calculates how much current TS changes
    compared to previous historical TS. It includes method to calculate t-scores, upper bound, lower bound, etc., for
    Statsig Detector model to detect significant change.

    Attributes:
        current: ChangePointInterval. The TS interval we'd like to detect.
        previous: ChangePointInterval. The historical TS interval, which we compare the current TS interval with.
        method: str, default value is "fdr_bh".
        skip_rescaling: bool, default value is False. For multi-variate TS, we need to rescale p-values so that alpha is still the threshold for rejection.
                    For Statsig detector, when a given TS (except historical part) is longer than max_split_ts_length,
                    we will transform this long univariate TS into a multi-variate TS and then use multistatsig detector instead.
                    In this case, we should skip rescaling p-values.
        use_corrected_scores: bool, default value is False, using original t-scores or correct t-scores.
        min_perc_change: float, minimum percentage change, for a non zero score. Score will be clipped to zero if the absolute value of the percentage chenge is less than this value
    """

    upper: Optional[Union[float, np.ndarray]]
    lower: Optional[Union[float, np.ndarray]]
    _t_score: Optional[Union[float, np.ndarray]]
    _p_value: Optional[Union[float, np.ndarray]]
    num_series: int

    def __init__(
        self,
        current: ChangePointInterval,
        previous: ChangePointInterval,
        method: str = "fdr_bh",
        skip_rescaling: bool = False,
        use_corrected_scores: bool = False,
        min_perc_change: float = 0.0,
    ) -> None:
        self.current = current
        self.previous = previous

        self.upper = None
        self.lower = None
        self._t_score = None
        self._p_value = None
        self.alpha = 0.05
        self.method = method
        self.num_series = self.current.num_series

        # If we'd like skip rescaling p-values for multivariate timeseires data
        self.skip_rescaling = skip_rescaling

        # 2 t scores strategies
        self.use_corrected_scores = use_corrected_scores
        self.min_perc_change = min_perc_change

    @property
    def ratio_estimate(self) -> Union[float, np.ndarray]:
        # pyre-ignore[6]: Expected float for 1st positional only parameter to call float.__truediv__ but got Union[float, np.ndarray].
        return self.current.mean_val / self.previous.mean_val

    @property
    def perc_change(self) -> float:
        return (self.ratio_estimate - 1.0) * 100.0

    @property
    def perc_change_upper(self) -> float:
        if self.upper is None:
            self._delta_method()
        return (cast(Union[float, np.ndarray], self.upper) - 1) * 100.0

    @property
    def perc_change_lower(self) -> float:
        if self.lower is None:
            self._delta_method()
        return (cast(Union[float, np.ndarray], self.lower) - 1) * 100.0

    @property
    def direction(self) -> Union[str, ArrayLike]:
        if self.num_series > 1:
            return np.vectorize(lambda x: "up" if x > 0 else "down")(self.perc_change)
        elif self.perc_change > 0.0:
            return "up"
        else:
            return "down"

    @property
    def stat_sig(self) -> Union[bool, ArrayLike]:
        if self.upper is None:
            self._delta_method()
        if self.num_series > 1:
            return np.array(
                [
                    False if cast(np.ndarray, self.upper)[i] > 1.0 and cast(np.ndarray, self.lower)[i] < 1 else True
                    for i in range(self.current.num_series)
                ]
            )
        # not stat sig e.g. [0.88, 1.55]
        return not (
            cast(Union[float, np.ndarray], self.upper) > 1.0 and cast(Union[float, np.ndarray], self.lower) < 1.0
        )

    @property
    def score(self) -> float:
        if self._t_score is None:
            self._ttest()

        t_score = self._t_score

        if self.num_series == 1:
            if np.abs(self.perc_change) < self.min_perc_change:
                t_score = 0.0
        else:
            t_score = np.where(np.abs(self.perc_change) < self.min_perc_change, 0, t_score)

        return cast(float, t_score)

    @property
    def p_value(self) -> float:
        if self._p_value is None:
            self._ttest()
        return cast(float, self._p_value)

    @property
    def mean_previous(self) -> Union[float, np.ndarray]:
        return self.previous.mean_val

    @property
    def mean_difference(self) -> Union[float, np.ndarray]:
        # pyre-ignore[6]: Expected `float` for 1st param but got `Union[float,
        #  np.ndarray]`.
        _mean_diff = self.current.mean_val - self.previous.mean_val
        return _mean_diff

    @property
    def ci_upper(self) -> float:
        sp_mean = self._pooled_stddev()
        df = self._get_df()

        # the minus sign here is non intuitive.
        # this is because, for example, t.ppf(0.025, 30) ~ -1.96
        _ci_upper = self.previous.mean_val - t.ppf(self.alpha / 2, df) * sp_mean

        return _ci_upper

    @property
    def ci_lower(self) -> float:
        sp_mean = self._pooled_stddev()
        df = self._get_df()
        # the plus sign here is non-intuitive. See comment
        # above
        _ci_lower = self.previous.mean_val + t.ppf(self.alpha / 2, df) * sp_mean

        return _ci_lower

    def _get_df(self) -> float:
        """
        degree of freedom of t-test
        """
        n_1 = len(self.previous)
        n_2 = len(self.current)
        df = n_1 + n_2 - 2

        return df

    def _pooled_stddev(self) -> float:
        """
        This calculates the pooled standard deviation for t-test
        as defined in https://online.stat.psu.edu/stat500/lesson/7/7.3/7.3.1/7.3.1.1
        """

        s_1_sq = self.previous.variance_val
        s_2_sq = self.current.variance_val
        n_1 = len(self.previous)
        n_2 = len(self.current)

        if n_1 == 0 or n_2 == 0:
            return 0.0

        # pyre-ignore[58]: * is not supported for operand types int and Union[float, np.ndarray].
        s_p = np.sqrt(((n_1 - 1) * s_1_sq + (n_2 - 1) * s_2_sq) / (n_1 + n_2 - 2))

        if not self.use_corrected_scores:
            return s_p

        # based on the definition of t-test, we should return s_p_mean
        s_p_mean = s_p * np.sqrt((1.0 / n_1) + (1.0 / n_2))
        return s_p_mean

    def _ttest_manual(self) -> Tuple[float, float]:
        """
        scipy's t-test gives nan when one of the arrays has a
        size of 1.
        To repro, run:
        >>> ttest_ind(np.array([1,2,3,4]), np.array([11]), equal_var=True, nan_policy='omit')
        This is implemented to fix this issue
        """
        sp_mean = self._pooled_stddev()
        df = self._get_df()

        # pyre-ignore[6]: Expected float for 1st positional only parameter to call float.__sub__ but got Union[float, np.ndarray].
        t_score = (self.current.mean_val - self.previous.mean_val) / sp_mean
        p_value = t.sf(np.abs(t_score), df) * 2  # sf = 1 - cdf

        return t_score, p_value

    def _ttest(self) -> None:
        if self.num_series > 1:
            self._ttest_multivariate()
            return

        n_1 = len(self.previous)
        n_2 = len(self.current)

        # if both control and test have one value
        # then using a t test does not make any sense
        if n_1 == 1 and n_2 == 1:
            self._t_score = np.inf
            self._p_value = 0.0

        # when sample size is 1, scipy's t test gives nan,
        # hence we separately handle this case
        # if n_1 == 1 or n_2 == 1:
        #     self._t_score, self._p_value = self._ttest_manual()
        # else:
        #     self._t_score, self._p_value = ttest_ind(
        #         current_data, prev_data, equal_var=True, nan_policy='omit'
        #     )

        # Always use ttest_manual because we changed the std to not include
        # np.sqrt((1. / n_1) + (1./ n_2))
        self._t_score, self._p_value = self._ttest_manual()

    def _ttest_multivariate(self) -> None:
        num_series = self.num_series
        p_value_start = np.zeros(num_series)
        t_value_start = np.zeros(num_series)

        n_1 = len(self.previous)
        n_2 = len(self.current)

        if n_1 == 1 and n_2 == 1:
            self._t_score = np.inf * np.ones(num_series)
            self._p_value = np.zeros(num_series)
            return
        elif n_1 == 1 or n_2 == 1:
            t_value_start, p_value_start = self._ttest_manual()
        else:
            current_data = self.current.data
            prev_data = self.previous.data
            if current_data is None or prev_data is None:
                raise ValueError("Interval data not set")
            for i in range(num_series):
                current_slice = current_data[:, i]
                prev_slice = prev_data[:, i]
                t_value_start[i], p_value_start[i] = ttest_ind(
                    current_slice, prev_slice, equal_var=True, nan_policy="omit"
                )

        # if un-scaled t_score and p_value are needed
        if self.skip_rescaling:
            self._p_value = p_value_start
            self._t_score = t_value_start
            return

        # The new p-values are the old p-values rescaled so that self.alpha is still the threshold for rejection
        _, self._p_value, _, _ = multitest.multipletests(p_value_start, alpha=self.alpha, method=self.method)
        self._t_score = np.zeros(num_series)
        # We are using a two-sided test here, so we take inverse_tcdf(self._p_value / 2) with df = len(self.current) + len(self.previous) - 2

        _t_score: np.ndarray = self._t_score
        _p_value: np.ndarray = cast(np.ndarray, self._p_value)
        for i in range(self.current.num_series):
            if t_value_start[i] < 0:
                _t_score[i] = t.ppf(_p_value[i] / 2, self._get_df())
            else:
                _t_score[i] = t.ppf(1 - _p_value[i] / 2, self._get_df())
        self._t_score = _t_score

    def _calc_cov(self) -> float:
        """
        Calculates the covariance of x and y
        """
        current = self.current.data
        previous = self.previous.data
        if current is None or previous is None:
            return np.nan
        n_min = min(len(current), len(previous))
        if n_min == 0:
            return np.nan

        current = current[-n_min:]
        previous = previous[-n_min:]

        # for multivariate TS data
        if self.num_series > 1:
            return np.asarray([np.cov(current[:, c], previous[:, c])[0, 1] / n_min for c in range(self.num_series)])

        return np.cov(current, previous)[0, 1] / n_min

    def _delta_method(self) -> None:
        test_mean = self.current.mean_val
        control_mean = self.previous.mean_val
        test_var = self.current.variance_val
        control_var = self.previous.variance_val

        n_test = len(self.current)
        n_control = len(self.previous)

        cov_xy = self._calc_cov()

        sigma_sq_ratio = (
            test_var / (n_test * (control_mean**2))
            - 2 * (test_mean * cov_xy) / (control_mean**3)
            + (control_var * (test_mean**2)) / (n_control * (control_mean**4))
        )
        # the signs appear flipped because norm.ppf(0.025) ~ -1.96
        self.lower = self.ratio_estimate + norm.ppf(self.alpha / 2) * np.sqrt(abs(sigma_sq_ratio))
        self.upper = self.ratio_estimate - norm.ppf(self.alpha / 2) * np.sqrt(abs(sigma_sq_ratio))


@dataclass
class ConfidenceBand:
    lower: TimeSeriesData
    upper: TimeSeriesData


class AnomalyResponse:

    key_mapping: List[str]
    num_series: int

    def __init__(
        self,
        scores: TimeSeriesData,
        confidence_band: Optional[ConfidenceBand],
        predicted_ts: Optional[TimeSeriesData],
        anomaly_magnitude_ts: TimeSeriesData,
        stat_sig_ts: Optional[TimeSeriesData],
    ) -> None:
        self.scores = scores
        self.confidence_band = confidence_band
        self.predicted_ts = predicted_ts
        self.anomaly_magnitude_ts = anomaly_magnitude_ts
        self.stat_sig_ts = stat_sig_ts

        self.key_mapping = []
        self.num_series = 1

        if not self.scores.is_univariate():
            self.num_series = len(scores.value.columns)
            self.key_mapping = list(scores.value.columns)

    def update(
        self,
        time: datetime,
        score: Union[float, ArrayLike],
        ci_upper: Union[float, ArrayLike],
        ci_lower: Union[float, ArrayLike],
        pred: Union[float, ArrayLike],
        anom_mag: Union[float, ArrayLike],
        stat_sig: Union[float, ArrayLike],
    ) -> None:
        """
        Add one more point and remove the last point
        """
        self.scores = self._update_ts_slice(self.scores, time, score)
        confidence_band = self.confidence_band
        if confidence_band is not None:
            self.confidence_band = ConfidenceBand(
                lower=self._update_ts_slice(confidence_band.lower, time, ci_lower),
                upper=self._update_ts_slice(confidence_band.upper, time, ci_upper),
            )

        predicted_ts = self.predicted_ts
        if predicted_ts is not None:
            self.predicted_ts = self._update_ts_slice(predicted_ts, time, pred)
        self.anomaly_magnitude_ts = self._update_ts_slice(self.anomaly_magnitude_ts, time, anom_mag)
        stat_sig_ts = self.stat_sig_ts
        if stat_sig_ts is not None:
            self.stat_sig_ts = self._update_ts_slice(stat_sig_ts, time, stat_sig)

    def _update_ts_slice(self, ts: TimeSeriesData, time: datetime, value: Union[float, ArrayLike]) -> TimeSeriesData:
        time = ts.time.iloc[1:].append(pd.Series(time, copy=False))
        time.reset_index(drop=True, inplace=True)
        if self.num_series == 1:
            value = ts.value.iloc[1:].append(pd.Series(value, copy=False))
            value.reset_index(drop=True, inplace=True)
            return TimeSeriesData(time=time, value=value)
        else:
            if isinstance(value, float):
                raise ValueError(f"num_series = {self.num_series} so value should have type ArrayLike.")
            value_dict = {}
            for i, value_col in enumerate(self.key_mapping):
                value_dict[value_col] = ts.value[value_col].iloc[1:].append(pd.Series(value[i], copy=False))
                value_dict[value_col].reset_index(drop=True, inplace=True)
            return TimeSeriesData(
                pd.DataFrame(
                    {
                        **{"time": time},
                        **{value_col: value_dict[value_col] for value_col in self.key_mapping},
                    },
                    copy=False,
                )
            )

    def inplace_update(
        self,
        time: datetime,
        score: Union[float, ArrayLike],
        ci_upper: Union[float, ArrayLike],
        ci_lower: Union[float, ArrayLike],
        pred: Union[float, ArrayLike],
        anom_mag: Union[float, ArrayLike],
        stat_sig: Union[float, ArrayLike],
    ) -> None:
        """
        Add one more point and remove the last point
        """
        self._inplace_update_ts(self.scores, time, score)
        cb = self.confidence_band
        if cb is not None:
            self._inplace_update_ts(cb.lower, time, ci_lower),
            self._inplace_update_ts(cb.upper, time, ci_upper)

        self._inplace_update_ts(self.predicted_ts, time, pred)
        self._inplace_update_ts(self.anomaly_magnitude_ts, time, anom_mag)
        self._inplace_update_ts(self.stat_sig_ts, time, stat_sig)

    def _inplace_update_ts(
        self,
        ts: Optional[TimeSeriesData],
        time: datetime,
        value: Union[float, ArrayLike],
    ) -> None:
        if ts is None:
            return
        if self.num_series == 1:
            ts.value.loc[ts.time == time] = value
        else:
            ts.value.loc[ts.time == time, :] = np.array(value, dtype=float)

    def get_last_n(self, N: int) -> AnomalyResponse:
        """
        returns the response for the last N days
        """
        cb = self.confidence_band
        pts = self.predicted_ts
        ssts = self.stat_sig_ts

        return AnomalyResponse(
            scores=self.scores[-N:],
            confidence_band=None
            if cb is None
            else ConfidenceBand(
                upper=cb.upper[-N:],
                lower=cb.lower[-N:],
            ),
            predicted_ts=None if pts is None else pts[-N:],
            anomaly_magnitude_ts=self.anomaly_magnitude_ts[-N:],
            stat_sig_ts=None if ssts is None else ssts[-N:],
        )

    def __str__(self) -> str:
        cb = self.confidence_band
        upper = None if cb is None else cb.upper.value.values
        lower = None if cb is None else cb.lower.value.values
        predicted = None if self.predicted_ts is None else self.predicted_ts.value.values
        statsig = None if self.stat_sig_ts is None else self.stat_sig_ts.value.values
        str_ret = f"""
        Time: {self.scores.time.values},
        Scores: {self.scores.value.values},
        Upper Confidence Bound: {upper},
        Lower Confidence Bound: {lower},
        Predicted Time Series: {predicted},
        stat_sig:{statsig}
        """

        return str_ret



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

""" Collection of methods that return default search spaces for their relevant models.

This module has a collection of functions. Each function is dedicated for a model. It
returns default search space for hyperparameter tuning that pertains to the model.
They are called by hyperparemeter tuning module, time_series_parameter_tuning.py.

"""


from typing import Any
from typing import Dict
from typing import List

import numpy as np


def get_default_arnet_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for arnet.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: input_size, output_size, batch_size.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "input_size",
            "type": "choice",
            "values": list(range(3, 14)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "output_size",
            "type": "choice",
            "values": list(range(3, 14)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "batch_size",
            "type": "choice",
            "values": list(range(5, 20)),
            "value_type": "int",
            "is_ordered": True,
        },
    ]


def get_default_stlf_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for stfl.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: method, m.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "method",
            "type": "choice",
            "value_type": "str",
            "values": ["linear", "quadratic", "theta"],
        },
        {
            "name": "m",
            "type": "choice",
            # The number of periods in this seasonality
            # (e.g. 7 periods for daily data would be used for weekly seasonality)
            "values": [4, 7, 10, 14, 24, 30],
            "value_type": "int",
            "is_ordered": True,
        },
    ]


def get_default_arima_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for arima.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: p, d, q.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "p",
            "type": "choice",
            "values": list(range(1, 6)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "d",
            "type": "choice",
            "values": list(range(1, 3)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "q",
            "type": "choice",
            "values": list(range(1, 6)),
            "value_type": "int",
            "is_ordered": True,
        },
    ]


def get_default_holtwinters_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for holtwinters.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: trend, damped, seasonal, seasonal_periods.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "trend",
            "type": "choice",
            "value_type": "str",
            "values": ["additive", "multiplicative"],
        },
        {
            "name": "damped",
            "type": "choice",
            "value_type": "bool",
            "values": [True, False],
        },
        {
            "name": "seasonal",
            "type": "choice",
            "value_type": "str",
            "values": ["additive", "multiplicative"],
        },
        {
            "name": "seasonal_periods",
            "type": "choice",
            # The number of periods in this seasonality
            # (e.g. 7 periods for daily data would be used for weekly seasonality)
            "values": [4, 7, 10, 14, 24, 30],
            "value_type": "int",
            "is_ordered": True,
        },
    ]


def get_default_sarima_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for sarima.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: p, d, q, seasonal_order, trend.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "p",
            "type": "choice",
            "values": list(range(1, 6)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "d",
            "type": "choice",
            "values": list(range(1, 3)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "q",
            "type": "choice",
            "values": list(range(1, 6)),
            "value_type": "int",
            "is_ordered": True,
        },
        {
            "name": "seasonal_order",
            "type": "choice",
            "values": [
                (1, 0, 1, 7),
                (1, 0, 2, 7),
                (2, 0, 1, 7),
                (2, 0, 2, 7),
                (1, 1, 1, 7),
                (0, 1, 1, 7),
            ],
            # Note: JSON representation must be 'int', 'float', 'bool' or 'str'.
            # so we use 'str' here instead of 'Tuple'
            # when doing HPT, we need to convert it back to tuple
            "value_type": "str",
        },
        {
            "name": "trend",
            "type": "choice",
            "values": ["n", "c", "t", "ct"],
            "value_type": "str",
        },
    ]


def get_default_theta_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for theta.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: m.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "m",
            "type": "choice",
            # Number of observations before the seasonal pattern repeats
            # e.g. m=12 for montly data with yearly seasonality
            "values": list(range(1, 31)),
            "value_type": "int",
            "is_ordered": True,
        },
    ]


def get_default_var_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for var.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: To be filled.

    Args:
        N/A

    Returns:
        As described above

    Raises:
        NotImplementedError: The method is to be implemented in the future. This
        error will then be removed.
    """

    # TODO: remove raise error, then implement the default parameter
    # space definition.
    raise NotImplementedError("get_parameter_search_space() method has not been implemented for " "VAR model.")


def get_default_lightgbm_parameter_search_space() -> List[Dict[str, Any]]:
    """Generates default search space as a list of dictionaries and returns it for lightgbm model.

    Each dictionary in the list corresponds to a hyperparameter, having properties
    defining that hyperparameter. Properties are name, type, value_type, values,
    is_ordered. Hyperparameters that are included: n_estimators, max_depth, learning_rate,
    min_split_gain, num_leaves, reg_alpha, subsample_for_bin

    Args:
        N/A

    Returns:
        As described above

    Raises:
        N/A
    """

    return [
        {
            "name": "n_estimators",
            "type": "choice",
            "value_type": "int",
            "values": list(np.arange(100, 800, 50)),
            "is_ordered": True,
        },
        {
            "name": "max_depth",
            "type": "choice",
            "value_type": "int",
            "values": list(np.arange(5, 20, 2)),
            "is_ordered": True,
        },
        {
            "name": "learning_rate",
            "type": "choice",
            "value_type": "float",
            "values": list(np.logspace(1e-3, 0.8, endpoint=True)),
            "is_ordered": True,
        },
        {
            "name": "min_split_gain",
            "type": "choice",
            "value_type": "float",
            "values": list(np.arange(1e-4, 1e-1, 0.5e-3)),
            "is_ordered": True,
        },
        {
            "name": "num_leaves",
            "type": "choice",
            "value_type": "int",
            "values": list(np.arange(2**5, 2**11, 128)),
            "is_ordered": True,
        },
    ]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

try:
    from . import metalearner_hpt  # noqa
except ImportError:
    import logging

    logging.warning("kats.models.metalearner.metalearner_hpt requires torch be installed")
from . import metalearner_modelselect  # noqa  # noqa
from . import metalearner_predictability



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from unittest import TestCase

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.detectors.outlier_detector import OutlierDetectorModel


class TestOutlierDetectorModel(TestCase):
    def setUp(self) -> None:
        self.data = load_air_passengers(return_ts=False)
        self.outlier_detector = OutlierDetectorModel()

    def test_response_shape_for_single_series(self) -> None:
        single_ts = TimeSeriesData(self.data)
        response_single_ts = self.outlier_detector.fit_predict(data=single_ts, historical_data=None)

        self.assertEqual(response_single_ts.scores.time.shape, single_ts.time.shape)

        self.assertEqual(response_single_ts.scores.value.shape, single_ts.value.shape)

        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `value`.
            response_single_ts.predicted_ts.value.shape,
            single_ts.value.shape,
        )

    def test_response_shape_for_multi_series(self) -> None:
        data = self.data.copy()
        data["y_2"] = data["y"]
        multi_ts = TimeSeriesData(data)
        response_multi_ts = self.outlier_detector.fit_predict(data=multi_ts, historical_data=None)

        self.assertEqual(response_multi_ts.scores.time.shape, multi_ts.time.shape)

        self.assertEqual(response_multi_ts.scores.value.shape, multi_ts.value.shape)

        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `value`.
            response_multi_ts.predicted_ts.value.shape,
            multi_ts.value.shape,
        )

    def test_response_shape_with_historical_data(self) -> None:
        single_ts = TimeSeriesData(self.data)
        historical_ts = TimeSeriesData(self.data)
        single_ts.time = single_ts.time + pd.tseries.offsets.DateOffset(months=len(self.data))
        response = self.outlier_detector.fit_predict(single_ts, historical_ts)

        self.assertTrue(np.array_equal(response.scores.time, single_ts.time))

    def test_pmm_use_case(self) -> None:
        random.seed(100)
        time_unit = 86400
        hist_data_time = [x * time_unit for x in range(0, 28)]
        data_time = [x * time_unit for x in range(28, 35)]

        hist_data_value = [random.normalvariate(100, 10) for _ in range(0, 28)]
        data_value = [random.normalvariate(130, 10) for _ in range(28, 35)]

        hist_ts = TimeSeriesData(
            time=pd.Series(hist_data_time),
            value=pd.Series(hist_data_value),
            use_unix_time=True,
            unix_time_units="s",
        )
        data_ts = TimeSeriesData(
            time=pd.Series(data_time),
            value=pd.Series(data_value),
            use_unix_time=True,
            unix_time_units="s",
        )

        response_with_historical_data = self.outlier_detector.fit_predict(data=data_ts, historical_data=hist_ts)
        self.assertEqual(response_with_historical_data.scores.value.shape, data_ts.value.shape)
        response_wo_historical_data = self.outlier_detector.fit_predict(data=hist_ts)
        self.assertEqual(response_wo_historical_data.scores.value.shape, hist_ts.value.shape)

    def test_remover_usecase(self) -> None:
        # manually add outlier on the date of '1950-12-01'
        self.data.loc[self.data.time == "1950-12-01", "y"] *= 5
        # manually add outlier on the date of '1959-12-01'
        self.data.loc[self.data.time == "1959-12-01", "y"] *= 4

        single_ts = TimeSeriesData(self.data)

        response_with_interpolate = self.outlier_detector.fit_predict(single_ts, historical_ts=None, interpolate=True)

        response_with_no_interpolate = self.outlier_detector.fit_predict(
            data=single_ts, historical_data=None, interpolate=False
        )

        # When no iterpolate argument is given by default it is taking False
        response_with_default_interpolate = self.outlier_detector.fit_predict(data=single_ts, historical_data=None)

        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `time`.
            response_with_interpolate.predicted_ts.time.shape,
            single_ts.time.shape,
        )
        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `value`.
            response_with_interpolate.predicted_ts.value.shape,
            single_ts.value.shape,
        )

        self.assertEqual(response_with_no_interpolate.predicted_ts.time.shape, single_ts.time.shape)
        self.assertEqual(response_with_no_interpolate.predicted_ts.value.shape, single_ts.value.shape)

        self.assertEqual(
            response_with_default_interpolate.predicted_ts.time.shape,
            single_ts.time.shape,
        )
        self.assertEqual(
            response_with_default_interpolate.predicted_ts.value.shape,
            single_ts.value.shape,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import numpy as np
from sklearn.metrics import mean_squared_error as mse
from sklearn.metrics import mean_squared_log_error as msle

from kats.compat import compat

version: compat.Version = compat.Version("sklearn")


def mean_squared_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    multioutput: str = "uniform_average",
    squared: bool = True,
) -> float:
    # sklearn >= 1.0 expects args beyond the first two to be passed as named kwargs
    return mse(
        y_true,
        y_pred,
        sample_weight=sample_weight,
        multioutput=multioutput,
        squared=squared,
    )


def mean_squared_log_error(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    sample_weight: Optional[np.ndarray] = None,
    multioutput: str = "uniform_average",
    squared: bool = True,
) -> float:
    if version <= "0.24":
        result = msle(y_true, y_pred, sample_weight=sample_weight, multioutput=multioutput)
        if not squared:
            result = np.sqrt(result)
    else:
        # sklearn >= 1.0 expects args beyond the first two to be passed as named kwargs
        # pyre-ignore
        result = msle(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
            squared=squared,
        )
    return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import unittest
from typing import Any
from typing import Dict
from typing import cast
from unittest import TestCase

import pandas as pd
from parameterized.parameterized import parameterized

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models.holtwinters import HoltWintersModel
from kats.models.holtwinters import HoltWintersParams
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_HW_1,  # first param combination results
)
from kats.tests.models.test_models_dummy_data import AIR_FCST_HW_2

pd_ver = float(re.findall("([0-9]+\\.[0-9]+)\\..*", pd.__version__)[0])

ADD_PARAMS = HoltWintersParams(
    trend="add",
    seasonal="add",
    seasonal_periods=7,
)
MUL_PARAMS = HoltWintersParams(
    trend="mul",
    seasonal="mul",
    seasonal_periods=7,
)
TS: TimeSeriesData = cast(TimeSeriesData, load_air_passengers())


class HoltWintersModelTest(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly_addiitive",
                TS,
                "MS",
                ADD_PARAMS,
                {"freq": "MS", "include_history": True},
                AIR_FCST_HW_1,
            ],
            [
                "monthly_multiplicative",
                TS,
                "MS",
                MUL_PARAMS,
                {"alpha": 0.9},
                AIR_FCST_HW_2,
            ],
        ]
    )
    def test_fcst(
        self,
        name: str,
        ts: TimeSeriesData,
        freq: str,
        params: HoltWintersParams,
        predict_args: Dict[str, Any],
        truth: pd.DataFrame,
    ) -> None:
        params.validate_params()

        # Fit forecast
        m = HoltWintersModel(ts, params)
        m.fit()
        res = m.predict(steps=30, **predict_args)

        # Test result
        assert_frame_equal(truth, res, check_less_precise=0, rtol=1)

    def test_invalid_params(self) -> None:
        self.assertRaises(
            ValueError,
            HoltWintersParams,
            trend="random_trend",
        )

        self.assertRaises(
            ValueError,
            HoltWintersParams,
            seasonal="random_seasonal",
        )

        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        TSData_multi = TimeSeriesData(DATA_multi)

        params = HoltWintersParams()

        self.assertRaises(
            ValueError,
            HoltWintersModel,
            data=TSData_multi,
            params=params,
        )

    def test_exec_plot(self) -> None:
        # Fit forecast
        m = HoltWintersModel(TS, ADD_PARAMS)
        m.fit()
        _ = m.predict(steps=2, freq="MS")

        # Test plotting
        m.plot()

    def test_name(self) -> None:
        m = HoltWintersModel(TS, ADD_PARAMS)
        self.assertEqual(m.__str__(), "HoltWinters")

    def test_search_space(self) -> None:
        m = HoltWintersModel(TS, ADD_PARAMS)
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "trend",
                    "type": "choice",
                    "value_type": "str",
                    "values": ["additive", "multiplicative"],
                },
                {
                    "name": "damped",
                    "type": "choice",
                    "value_type": "bool",
                    "values": [True, False],
                },
                {
                    "name": "seasonal",
                    "type": "choice",
                    "value_type": "str",
                    "values": ["additive", "multiplicative"],
                },
                {
                    "name": "seasonal_periods",
                    "type": "choice",
                    "values": [4, 7, 10, 14, 24, 30],
                    "value_type": "int",
                    "is_ordered": True,
                },
            ],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import ast
import logging
import math
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import pandas as pd

from kats.consts import TimeSeriesData
from kats.models.metalearner.metalearner_modelselect import MetaLearnModelSelect
from kats.tsfeatures.tsfeatures import TsFeatures

NUM_SECS_IN_DAY: int = 3600 * 24
PARAMS_TO_SCALE_DOWN = {"historical_window", "scan_window"}
MIN_EXAMPLES = 30


def get_ts_features(ts: TimeSeriesData) -> Dict[str, float]:
    """
    Extract TSFeatures for the input time series.
    """
    # Run Kats TsFeatures
    ts_features = TsFeatures(hw_params=False)
    feats = ts_features.transform(ts)

    # Rounding features
    features = {}
    assert isinstance(feats, dict)
    for feature_name, feature_val in feats.items():
        if not math.isnan(feature_val):
            feature_val = format(feature_val, ".4f")
        features[feature_name] = feature_val

    return features


def change_dtype(d: Dict[str, Any]) -> Dict[str, float]:
    for elm in d:
        d[elm] = float(d[elm])
    return d


def change_str_to_dict(x: Union[str, Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    if isinstance(x, dict):
        return x
    elif isinstance(x, str):
        try:
            dict_str = ast.literal_eval(x)
            return dict_str
        except Exception as e:
            print(e)
    else:
        raise ValueError(f"type is {type(x)}")


def metadata_detect_preprocessor(
    rawdata: pd.DataFrame,
    params_to_scale_down: Set[str] = PARAMS_TO_SCALE_DOWN,
) -> List[Dict[str, Any]]:

    rawdata["features"] = rawdata["features"].map(change_str_to_dict)
    rawdata["features"] = rawdata["features"].map(change_dtype)
    rawdata["hpt_res"] = rawdata["hpt_res"].map(change_str_to_dict)

    table = [
        {
            "hpt_res": rawdata["hpt_res"].iloc[i],
            "features": rawdata["features"].iloc[i],
            "best_model": rawdata["best_model"].iloc[i],
        }
        for i in range(len(rawdata))
    ]

    for ts_data in table:
        for hpt_vals in ts_data["hpt_res"].values():
            params = hpt_vals[0]
            for param in params.keys():
                if param in params_to_scale_down:
                    params[param] = params[param] / NUM_SECS_IN_DAY

    return table


class MetaDetectModelSelect(object):
    """
    Meta-learner framework on detection model selection.
    This framework uses classification algorithms to recommend suitable detection models.
    For training, it uses time series features as inputs and the best detection models as labels.
    For prediction, it takes time series or time series features as inputs to predict the most suitable detection model.

    Attributes:
        metadata: pd.DataFrame;
            A list of dictionaries representing the meta-data of time series
            (e.g., the meta-data generated by Get_MetaLearn_Detect_Inputs object).
            Each dictionary d must contain at least 3 components: 'hpt_res', 'features' and 'best_model'.
            d['hpt_res'] represents the best hyper-parameters for each candidate model and the corresponding errors;
            d['features'] are time series features, and d['best_model'] is a string representing the best candidate model
            of the corresponding time series data.

    Sample Usage:
        >>> mdms = MetaDetectModelSelect(data)
        >>> mdms.train(n_trees=200, test_size=0.1, eval_method='mean') # Train a meta-learner model selection model.
        >>> mdms.report_metrics() # present training results
        >>> mdms.predict(TSdata) # Predict the most suitable detection model.
        >>> mdms.fit_results() # present fitting results
        >>> mdms.pred_by_feature(eval_df) # Predict the most suitable detection model by features.
    """

    def __init__(
        self,
        metadata_df: pd.DataFrame,
        params_to_scale_down: Set[str] = PARAMS_TO_SCALE_DOWN,
    ) -> None:
        if not isinstance(metadata_df, pd.DataFrame):
            msg = "Dataset is not in form of a dataframe!"
            logging.error(msg)
            raise ValueError(msg)

        if len(metadata_df) <= MIN_EXAMPLES:
            msg = "Dataset is too small to train a meta learner!"
            logging.error(msg)
            raise ValueError(msg)

        expected_cols = ["hpt_res", "features", "best_model"]
        for col in expected_cols:
            if col not in metadata_df:
                msg = f"Missing column {col}, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

        self.metadata_df: pd.DataFrame = metadata_df
        self.results: Dict[str, Any] = {}
        self.mlms: Optional[MetaLearnModelSelect] = None
        self.params_to_scale_down: Set[str] = params_to_scale_down

    def _preprocess(self) -> List[Dict[str, Any]]:
        # prepare the training data
        # Create training data table
        return metadata_detect_preprocessor(self.metadata_df, self.params_to_scale_down)

    def train(
        self,
        method: str = "RandomForest",
        eval_method: str = "mean",
        test_size: float = 0.1,
        n_trees: int = 500,
        n_neighbors: int = 5,
    ) -> Dict[str, Any]:
        """
        Call the train() method of MetaLearnModelSelect, which returns
        {
            "fit_error": fit_error,
            "pred_error": pred_error,
            "clf_accuracy": metrics.accuracy_score(y_test, y_pred),
        }
        """
        self.mlms = MetaLearnModelSelect(self._preprocess())
        self.results = self.mlms.train(
            method=method,
            eval_method=eval_method,
            test_size=test_size,
            n_trees=n_trees,
            n_neighbors=n_neighbors,
        )
        return self.results

    def report_metrics(self) -> pd.DataFrame:
        # report the summary, as in the notebook N1154788
        if self.results is None:
            self.results = self.train()

        summary = pd.DataFrame([self.results["fit_error"], self.results["pred_error"]], copy=False)
        summary["type"] = ["fit_error", "pred_error"]
        summary["error_metric"] = "Inverted F-score"
        return summary

    def predict(self, ts: TimeSeriesData) -> str:
        # extract features from the ts
        feature_list = [get_ts_features(ts)]
        feature_list = [change_dtype(x) for x in feature_list]
        feature_df = pd.DataFrame(feature_list)

        if self.mlms is None:
            msg = "Please train a classifier first."
            logging.error(msg)
            raise ValueError(msg)
        algo_list = self.mlms.pred_by_feature(feature_df.values)
        return algo_list[0]

    def fit_results(self) -> pd.DataFrame:
        """
        Present fitting results
        """
        # extract features from the dataframe
        feature_list = [
            change_str_to_dict(self.metadata_df["features"].iloc[i]) for i in range(len(self.metadata_df.features))
        ]

        # pyre-fixme[6]
        feature_list = [change_dtype(x) for x in feature_list]
        feature_df = pd.DataFrame(feature_list)

        if self.mlms is None:
            msg = "Please train a classifier first."
            logging.error(msg)
            raise ValueError(msg)
        algo_list = self.mlms.pred_by_feature(feature_df.values)

        label_df = pd.DataFrame({"best_model": algo_list})

        return label_df

    def pred_by_feature(self, eval_df: pd.DataFrame) -> pd.DataFrame:
        """
        Extract features from the dataframe
        eval_df has one columns: features -> dict
        """
        feature_list = [change_str_to_dict(eval_df["features"].iloc[i]) for i in range(len(eval_df.features))]

        # pyre-fixme[6]
        feature_list = [change_dtype(x) for x in feature_list]
        feature_df = pd.DataFrame(feature_list)

        assert self.mlms is not None
        algo_list = self.mlms.pred_by_feature(feature_df.values)

        label_df = pd.DataFrame({"best_model": algo_list})

        return label_df



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import arima  # noqa  # noqa  # noqa  # noqa  # noqa  # noqa
from . import bayesian_var
from . import ensemble
from . import harmonic_regression
from . import holtwinters
from . import linear_model

try:
    from . import lstm  # noqa
except ImportError:
    import logging

    logging.warning("kats.models.lstm not available (requires torch)")
from . import metalearner
from . import model
from . import nowcasting
from . import quadratic_model
from . import reconciliation
from . import sarima
from . import stlf
from . import theta
from . import var



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
This module implements the multivariate Outlier Detection algorithm as a Detector Model.
"""
import json
from typing import Any
from typing import Optional

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.detectors.detector import DetectorModel
from kats.detectors.detector_consts import AnomalyResponse
from kats.detectors.detector_consts import ConfidenceBand
from kats.detectors.outlier import MultivariateAnomalyDetector
from kats.detectors.outlier import MultivariateAnomalyDetectorType
from kats.models.var import VARParams


class MultivariateAnomalyDetectorModel(DetectorModel):
    """Anamoly detection model based on outlier detector.

    A Detector Model detects outliers for multivariate time series

    Attributes:
        data: Input metrics TimeSeriesData
        params: Parameter class for multivariate VAR/ BVAR model
        training_days: num of days of data to use for initial training.
                    If less than a day, specify as fraction of a day
        model_type: The type of multivariate anomaly detector (currently 'BAYESIAN_VAR' and 'VAR' options available)
        serialized_model: Optional; json, representing data from a previously serialized model.
    """

    params: Params = VARParams(maxlags=2)
    training_days: float
    model_type: MultivariateAnomalyDetectorType = MultivariateAnomalyDetectorType.VAR
    model: Optional[MultivariateAnomalyDetector] = None

    def __init__(
        self,
        params: Params = params,
        training_days: float = 60.0,
        serialized_model: Optional[bytes] = None,
    ) -> None:
        if serialized_model:
            model_dict = json.loads(serialized_model)
            self.params = model_dict["params"]
            self.training_days = model_dict["training_days"]
        else:
            if params:
                self.params = params
            self.training_days = training_days

    def serialize(self) -> bytes:
        """Serialize the model into a json.

        Serialize the model into a json so it can be loaded later.

        Returns:
            json containing information of the model.
        """
        model_dict = {
            "params": self.params,
            "training_days": self.training_days,
            "model_type": self.model_type,
        }

        # model_dict = {"training_days": self.training_days, "model_type": self.model_type}
        return json.dumps(model_dict).encode("utf-8")

    def fit(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> None:
        """Fit MultivariateAnomalyDetector.

        Fit MultivariateAnomalyDetector using both data and historical_data (if provided).

        Args:
            data: TimeSeriesData on which detection is run.
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            None.
        """
        if historical_data is None:
            total_data = data
        else:
            historical_data.extend(data)
            total_data = historical_data

        self.model = MultivariateAnomalyDetector(
            data=total_data,
            params=self.params,
            training_days=self.training_days,
            model_type=self.model_type,
        )
        self.model.detector()

    def predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Get anomaly scores.

        Get anomaly scores only for data.

        Args:
            data: TimeSeriesData on which detection is run
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            AnomalyResponse object. The length of this obj.ect is same as data. The score property
            gives the score for anomaly.
        """
        if self.model is None:
            self.fit(data=data, historical_data=historical_data)

        assert self.model is not None
        output_scores_df = self.model.anomaly_score_df

        assert output_scores_df is not None
        output_scores_df = output_scores_df[output_scores_df.index >= data.time.min()]

        zeros = np.zeros(shape=(data.time.shape[0], output_scores_df.shape[1]))
        padding = np.empty(shape=[len(data) - output_scores_df.shape[0], output_scores_df.shape[1]])
        padding[:] = np.NaN
        padding = pd.DataFrame(padding, columns=output_scores_df.columns, copy=False)
        # all fields other than scores are left as TimeSeriesData with all zero values
        response = AnomalyResponse(
            scores=TimeSeriesData(
                time=data.time,
                value=pd.concat(
                    [padding.iloc[:, :-2], output_scores_df.iloc[:, :-2]],
                    ignore_index=True,
                    copy=False,
                ),
            ),
            confidence_band=ConfidenceBand(
                upper=TimeSeriesData(time=data.time, value=pd.DataFrame(zeros, copy=False)),
                lower=TimeSeriesData(time=data.time, value=pd.DataFrame(zeros, copy=False)),
            ),
            predicted_ts=TimeSeriesData(time=data.time, value=pd.DataFrame(zeros).iloc[:, :-2]),
            anomaly_magnitude_ts=TimeSeriesData(
                time=data.time,
                value=pd.concat(
                    [padding.iloc[:, -2], output_scores_df.iloc[:, -2]],
                    ignore_index=True,
                    copy=False,
                ),
            ),
            stat_sig_ts=TimeSeriesData(
                time=data.time,
                value=pd.concat(
                    [padding.iloc[:, -1], output_scores_df.iloc[:, -1]],
                    ignore_index=True,
                    copy=False,
                ),
            ),
        )

        return response

    def fit_predict(
        self,
        data: TimeSeriesData,
        historical_data: Optional[TimeSeriesData] = None,
        **kwargs: Any,
    ) -> AnomalyResponse:
        """Fit a model and return the anomaly scores.

        Return AnomalyResponse, when data is passed to it.

        Args:
            data: TimeSeriesData on which detection is run.
            historical_data: Optional; TimeSeriesData corresponding to history. History ends
                exactly where the data begins.

        Returns:
            AnomalyResponse object. The length of this object is same as data. The score property
            gives the score for anomaly.
        """
        self.fit(data=data, historical_data=historical_data)

        return self.predict(data=data)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from operator import attrgetter
from typing import Sequence
from typing import Tuple
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized
from sklearn.datasets import make_spd_matrix

from kats.consts import TimeSeriesData
from kats.detectors.robust_stat_detection import RobustStatChangePoint
from kats.detectors.robust_stat_detection import RobustStatDetector


class RobustStatTest(TestCase):
    def setUp(self) -> None:
        self.random_seed = 10
        np.random.seed(self.random_seed)
        self.dates = pd.Series(pd.date_range("2019-01-01", "2019-03-01"))

        # Initialize the change_points
        df_noregress = pd.DataFrame({"no_change": [math.sin(i) for i in range(60)]})

        df_increase = pd.DataFrame({"increase": [math.sin(i) if i < 41 else math.sin(i) + 17 for i in range(60)]})

        df_decrease = pd.DataFrame({"decrease": [math.sin(i) if i < 23 else math.sin(i) - 25 for i in range(60)]})

        df_spike_pos = pd.DataFrame({"spike": [math.sin(i) if i != 27 else 30 * math.sin(i) for i in range(60)]})

        df_spike_neg = pd.DataFrame({"spike": [math.sin(i) if i != 27 else -30 * math.sin(i) for i in range(60)]})

        def _create_change_pt(
            df: pd.DataFrame,
        ) -> Tuple[RobustStatDetector, Sequence[RobustStatChangePoint]]:
            df["time"] = self.dates
            timeseries = TimeSeriesData(df)
            detector = RobustStatDetector(timeseries)
            change_points = detector.detector()
            return detector, change_points

        self.d_noregress, self.change_points_noregress = _create_change_pt(df_noregress)
        self.d_increase, self.change_points_increase = _create_change_pt(df_increase)
        self.d_decrease, self.change_points_decrease = _create_change_pt(df_decrease)
        self.d_spike_pos, self.change_points_spike_pos = _create_change_pt(df_spike_pos)
        self.d_spike_neg, self.change_points_spike_neg = _create_change_pt(df_spike_neg)

        # Time series without name
        n = 10
        time = pd.Series(pd.date_range(start="2018-01-01", periods=n, freq="D"))
        value = pd.Series(np.arange(n))
        ts = TimeSeriesData(time=time, value=value)

        self.d_ts_without_name = RobustStatDetector(ts)
        self.change_points_ts_without_name = self.d_ts_without_name.detector()

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["change_points_noregress",
    #  0], ["change_points_increase", 1], ["change_points_decrease", 1],
    #  ["change_points_spike_pos", 2], ["change_points_spike_neg", 2]])`.
    @parameterized.expand(
        [
            ["change_points_noregress", 0],
            ["change_points_increase", 1],
            ["change_points_decrease", 1],
            ["change_points_spike_pos", 2],
            ["change_points_spike_neg", 2],
        ]
    )
    def test_length(self, change_points: str, ans: int) -> None:
        self.assertEqual(len(attrgetter(change_points)(self)), ans)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["d_noregress", "change_points_noregress"],
            ["d_increase", "change_points_increase"],
            ["d_decrease", "change_points_decrease"],
            ["d_spike_pos", "change_points_spike_pos"],
            ["d_spike_neg", "change_points_spike_neg"],
            ["d_ts_without_name", "change_points_ts_without_name"],
        ]
    )
    def test_plot(self, detector: str, change_points: str) -> None:
        attrgetter(detector)(self).plot(attrgetter(change_points)(self))

    def test_raise_error(self) -> None:
        D = 10
        mean1 = np.ones(D)
        mean2 = mean1 * 2
        sigma = make_spd_matrix(D, random_state=self.random_seed)

        df_increase = pd.DataFrame(
            np.concatenate(
                [
                    np.random.multivariate_normal(mean1, sigma, 60),
                    np.random.multivariate_normal(mean2, sigma, 30),
                ]
            )
        )

        df_increase["time"] = self.dates

        timeseries_multi = TimeSeriesData(df_increase)
        with self.assertRaises(ValueError):
            RobustStatDetector(timeseries_multi)
